# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module Unix
"""

from unittest import mock
import os
import random
import re
import subprocess
import tempfile
import unittest
from xml.dom.minidom import parseString

from tests import common
from bleachbit import logger
from bleachbit.FileUtilities import children_in_directory, exe_exists
from bleachbit.VFS import ListVFS
from bleachbit.Unix import (
    _is_broken_xdg_desktop_application,
    apt_autoclean,
    apt_autoremove,
    clear_snapd_cache,
    dnf_autoremove,
    dnf_clean,
    find_available_locales,
    find_best_locale,
    get_apt_size,
    get_distribution_name_version_distro,
    get_distribution_name_version_os_release,
    get_distribution_name_version_platform_freedesktop,
    get_distribution_name_version,
    get_purgeable_locales,
    get_trash_paths,
    is_broken_xdg_desktop,
    is_unix_display_protocol_wayland,
    journald_clean,
    JOURNALD_REGEX,
    LocaleCleanerPath,
    Locales,
    pacman_cache,
    root_is_not_allowed_to_X_session,
    snapd_is_active,
    snap_disabled_clean,
    snap_disabled_preview,
    snap_parse_list,
    rotated_logs,
    run_cleaner_cmd,
    wine_to_linux_path,
    yum_clean,
)


class FakeConfig:
    def __init__(self, options):
        self.options = options

    def has_option(self, section, option):
        return section in self.options and option in self.options[section]

    def get(self, section, option):
        return self.options[section][option]


class UnixTestCase(common.BleachbitTestCase):

    """Test case for module Unix"""

    def setUp(self):
        """Initialize unit tests"""
        self.locales = Locales()
        clear_snapd_cache()
        super(UnixTestCase, self).setUp()

    @common.skipIfWindows
    def test_apt(self):
        """Unit test for method apt_autoclean() and apt_autoremove()"""
        if 0 != os.geteuid() or not exe_exists('apt-get'):
            self.assertRaises(RuntimeError, apt_autoclean)
            self.assertRaises(RuntimeError, apt_autoremove)
        else:
            bytes_freed = apt_autoclean()
            self.assertIsInteger(bytes_freed)
            logger.debug('apt bytes cleaned %d', bytes_freed)
            bytes_freed = apt_autoremove()
            self.assertIsInteger(bytes_freed)
            logger.debug('apt bytes cleaned %d', bytes_freed)

    @common.skipIfWindows
    def test_find_available_locales(self):
        """Unit test for method find_available_locales()"""
        locales = find_available_locales()
        self.assertIsInstance(locales, list)
        for locale in locales:
            self.assertIsLanguageCode(locale)

    def test_find_available_locales_mock(self):
        """Unit test for method find_available_locales() with mock"""
        mock_locales = ['C', 'C.utf8', 'en_US.utf8',
                        'es_MX', 'es_MX.iso88591', 'es_MX.utf8', 'POSIX']
        with mock.patch('bleachbit.Unix.General.run_external') as mock_run_external:
            mock_run_external.return_value = (
                0, "\n".join(mock_locales) + "\n", "")
            locales = find_available_locales()
            self.assertEqual(locales, mock_locales)
            mock_run_external.assert_called_once_with(['locale', '-a'])

    @mock.patch('locale.getlocale')
    @mock.patch('bleachbit.Unix.find_available_locales')
    def test_find_best_locale(self, mock_find_available_locales, mock_getlocale):
        """Unit test for method find_best_locale()"""
        mock_find_available_locales.return_value = [
            'C',
            'C.utf8',
            'de_DE.utf8',
            'en_US.iso88591',
            'en_US.utf8',
            'es_MX',
            'es_MX.iso88591',
            'es_MX.utf8',
            'nds_DE.utf8',
            'POSIX',
        ]
        mock_getlocale.return_value = ('en_US', 'UTF-8')
        for locale in ('en', 'en_US', 'en_US.utf8'):
            self.assertEqual(find_best_locale(locale), 'en_US.UTF-8')

        mock_find_available_locales.assert_called()
        mock_getlocale.assert_called()

        for locale in ('de', 'de_DE', 'de_DE.utf8'):
            self.assertEqual(find_best_locale(locale), 'de_DE.utf8')

        # Test language with a three-letter code.
        for locale in ('nds', 'nds_DE', 'nds_DE.utf8'):
            self.assertEqual(find_best_locale(locale), 'nds_DE.utf8')

        # Reverse the list.
        mock_find_available_locales.return_value = mock_find_available_locales.return_value[
            ::-1]
        for locale in ('en', 'en_US', 'en_US.utf8'):
            self.assertEqual(find_best_locale(locale), 'en_US.UTF-8')

        # ISO-8859-1 is less preferred than UTF-8
        mock_find_available_locales.return_value.remove('en_US.utf8')
        mock_getlocale.return_value = ('es_MX', 'UTF-8')
        for locale in ('en', 'en_US'):
            self.assertEqual(find_best_locale(locale), 'en_US.iso88591')

        mock_getlocale.return_value = ('es_MX', 'UTF-8')
        mock_find_available_locales.return_value = ['C', 'C.utf8', 'en_US.utf8',
                                                    'es_MX.iso88591', 'es_MX.utf8', 'POSIX']
        for locale in ('es', 'es_MX', 'es_MX.utf8'):
            self.assertEqual(find_best_locale(locale), 'es_MX.UTF-8')

        self.assertEqual(find_best_locale('C'), 'C')
        self.assertEqual(find_best_locale(''), 'C')
        self.assertEqual(find_best_locale('POSIX'), 'POSIX')

        self.assertRaises(AssertionError, find_best_locale, None)
        self.assertRaises(AssertionError, find_best_locale, [])

    @unittest.skipUnless(exe_exists('apt-get'),
                         'skipping tests for unavailable apt-get')
    def test_get_apt_size(self):
        """Unit test for method get_apt_size()"""
        size = get_apt_size()
        self.assertIsInteger(size)
        self.assertGreaterEqual(size, 0)

    @common.skipIfWindows
    def test_get_distribution_name_version(self):
        """Unit test for method get_distribution_name_version()"""
        ret_real = get_distribution_name_version()
        ret_platform_freedesktop = get_distribution_name_version_platform_freedesktop()
        ret_distro = get_distribution_name_version_distro()
        ret_os_release = get_distribution_name_version_os_release()

        for ret in (ret_real, ret_platform_freedesktop, ret_distro, ret_os_release):
            if ret is None:
                continue
            self.assertIsInstance(ret, str)
            self.assertGreater(len(ret), 0)
            self.assertLess(len(ret), 100)
            self.assertIn(' ', ret)
        if ret_platform_freedesktop and ret_os_release:
            self.assertEqual(ret_platform_freedesktop, ret_os_release)
        if ret_distro and ret_os_release:
            self.assertEqual(ret_distro, ret_os_release)

        self.assertIsNotNone(ret_real)

    @common.skipIfWindows
    def test_is_broken_xdg_desktop_real(self):
        """Unit test for is_broken_xdg_desktop()

        Check it does not crash on real-world .desktop files.
        """
        menu_dirs = ['/usr/share/applications',
                     '~/.local/share/applications',
                     '~/snap/steam/common/.local/share/applications',
                     '/usr/share/autostart',
                     '/usr/share/gnome/autostart',
                     '/usr/share/gnome/apps',
                     '/usr/share/mimelnk',
                     '/usr/share/applnk-redhat/',
                     '/usr/local/share/applications/',
                     '/usr/share/ubuntu-wayland/']
        for dirname in menu_dirs:
            for filename in [fn for fn in children_in_directory(os.path.expanduser(dirname), False)
                             if fn.endswith('.desktop')
                             ]:
                self.assertIsInstance(is_broken_xdg_desktop(filename), bool)

    @common.skipIfWindows
    @mock.patch('bleachbit.Unix.logger.info')
    def test_is_broken_xdg_desktop_other(self, mock_logger):
        """Unit test for is_broken_xdg_desktop() using non-.desktop files"""
        system_dirs = ['/usr/bin', '/usr/lib', '/etc']
        filenames = []
        for dirname in system_dirs:
            for filename in children_in_directory(dirname, False):
                if filename.endswith('.desktop'):
                    continue
                filenames.append(filename)
        sample_size = min(1000, len(filenames))
        sampled_filenames = random.sample(filenames, sample_size)
        for filename in sampled_filenames:
            result = is_broken_xdg_desktop(filename)
            self.assertTrue(
                result, msg=f"Expected is_broken_xdg_desktop({filename}) to return True, but got {result}")
            mock_logger.assert_called()
            mock_logger.reset_mock()

    def test_is_broken_xdg_desktop_wine(self):
        """Unit test for certain Wine .desktop file

        https://github.com/bleachbit/bleachbit/issues/1756
        """
        if not os.getenv('PATH'):
            self.skipTest("PATH environment variable is not set")
        with tempfile.NamedTemporaryFile(mode='w', suffix='.desktop', prefix='bleachbit-xdg-') as tmp:
            tmp.write("""[Desktop Entry]
Name=What's New
Exec=env WINEPREFIX="/home/bem/.wine" wine C:\\\\windows\\\\command\\\\start.exe /Unix /home/bem/.wine/dosdevices/c:/users/Public/Start\\ Menu/Programs/Biet-O-Matic/What\\'s\\ New.lnk
Type=Application
StartupNotify=true
Path=/home/bem/.wine/dosdevices/c:/Program Files (x86)/Biet-O-Matic
Icon=6B19_WhatsNew.0""")
            tmp.flush()
            self.assertIsInstance(is_broken_xdg_desktop(tmp.name), bool)

    @mock.patch('bleachbit.FileUtilities.exe_exists')
    @common.skipIfWindows
    def test_is_broken_xdg_desktop_chrome(self, mock_exe_exists):
        """Unit test for certain Chrome .desktop file

        https://github.com/bleachbit/bleachbit/issues/1729
        """
        mock_exe_exists.return_value = True
        with tempfile.NamedTemporaryFile(mode='w', suffix='.desktop', prefix='bleachbit-xdg-') as tmp:
            tmp.write("""#!/usr/bin/env xdg-open
[Desktop Entry]
Version=1.0
Terminal=false
Type=Application
Name=Google News
Exec=/opt/google/chrome/google-chrome --profile-directory=Default --app-id=abcdefghijklmnopqrstuvwxy
Icon=chrome-abcdefghijklmnopqrstuvwxy-Default
StartupWMClass=crx_abcdefghijklmnopqrstuvwxy""")
            tmp.flush()
            self.assertFalse(is_broken_xdg_desktop(tmp.name))

    @common.skipIfWindows
    def test_is_broken_xdg_desktop_chatgpt(self):
        """Unit test for certain ChatGPT .desktop file

        Linux Mint made this file for a Web App.

        I changed Exec from vivaldi-stale to test without mock.

        https://github.com/bleachbit/bleachbit/issues/1729
        """
        with tempfile.NamedTemporaryFile(mode='w', suffix='.desktop', prefix='bleachbit-xdg-') as tmp:
            tmp.write("""[Desktop Entry]
Version=1.0
Name=ChatGPT
Comment=Web App
Exec=ls --app="https://chatgpt.com/" --class=WebApp-ChatGPT3878 --name=WebApp-ChatGPT3878
Terminal=false
X-MultipleArgs=false
Type=Application
Icon=/home/username/.local/share/icons/icons8-chatgpt-150.svg
Categories=GTK;Network;
MimeType=text/html;text/xml;application/xhtml_xml;
StartupWMClass=chatgpt.com
StartupNotify=true
X-WebApp-Browser=Vivaldi
X-WebApp-URL=https://chatgpt.com
X-WebApp-CustomParameters=
X-WebApp-Navbar=false
X-WebApp-PrivateWindow=false
X-WebApp-Isolated=false
Keywords=deepseek,chatgpt,gpt,gemini,ai
Hidden=false
NoDisplay=false
PrefersNonDefaultGPU=false""")
            tmp.flush()
            self.assertFalse(is_broken_xdg_desktop(tmp.name))


    @common.skipIfWindows
    def test_get_trash_paths(self):
        """Unit test for get_trash_paths()"""
        seen = []
        for p in get_trash_paths():
            seen.append(p)
            self.assertExists(p.path)
        self.assertEqual(len(seen), len(set(p.path for p in seen)), "Duplicate trash paths found")


    @common.skipIfWindows
    def test_desktop_valid_exe(self):
        """Unit test for .desktop file with valid Unix exe (not env)"""
        fake_config = FakeConfig({"Desktop Entry": {"Exec": "ls"}})
        if os.getenv('PATH'):
            result = _is_broken_xdg_desktop_application(
                fake_config, "foo.desktop")
            self.assertFalse(result)
        else:
            with self.assertRaises(RuntimeError):
                _is_broken_xdg_desktop_application(fake_config, "foo.desktop")

    def test_desktop_shlex_failure(self):
        """Unit test for .desktop file with shlex exception

        Malformed Exec values (unbalanced quotes) are undefined behavior per
        the XDG spec. The function returns False so the file is preserved
        rather than risking deletion of a working launcher.
        """
        fake_config = FakeConfig(
            {"Desktop Entry": {"Exec": "env ENVVAR=bar ls \"notepad.exe"}})
        result = _is_broken_xdg_desktop_application(
            fake_config, "foo.desktop")
        self.assertFalse(result)

    @mock.patch('bleachbit.FileUtilities.exe_exists')
    def test_desktop_quoted_exe_with_spaces(self, mock_exe_exists):
        """Regression test for #2118: AppImage .desktop with quoted Exec path

        Before the fix, the naive ``split(" ")[0]`` returned the executable
        path with its surrounding double-quotes still attached, so
        ``exe_exists`` reported it missing and the .desktop file was flagged
        for deletion.
        """
        mock_exe_exists.return_value = True
        fake_config = FakeConfig({"Desktop Entry": {
            "Exec": '"/home/user/Applications/AppManager" %u'}})
        result = _is_broken_xdg_desktop_application(
            fake_config, "com.github.AppManager.desktop")
        self.assertFalse(result)
        mock_exe_exists.assert_called_with('/home/user/Applications/AppManager')

    @mock.patch('bleachbit.FileUtilities.exe_exists')
    def test_desktop_quoted_exe_path_containing_spaces(self, mock_exe_exists):
        """Quoted Exec path that itself contains spaces must round-trip."""
        mock_exe_exists.return_value = True
        fake_config = FakeConfig({"Desktop Entry": {
            "Exec": '"/home/user/Applications/Converter NOW" %f'}})
        result = _is_broken_xdg_desktop_application(
            fake_config, "converter.desktop")
        self.assertFalse(result)
        mock_exe_exists.assert_called_with(
            '/home/user/Applications/Converter NOW')

    def test_desktop_empty_exec(self):
        """Whitespace-only Exec is treated as broken."""
        fake_config = FakeConfig({"Desktop Entry": {"Exec": "   "}})
        result = _is_broken_xdg_desktop_application(
            fake_config, "foo.desktop")
        self.assertTrue(result)

    @common.skipIfWindows
    def test_desktop_env_valid(self):
        """Unit test for .desktop file with valid env Exec"""
        fake_config = FakeConfig(
            {"Desktop Entry": {"Exec": "env FOO=bar ls /usr/bin"}})
        result = _is_broken_xdg_desktop_application(fake_config, "foo.desktop")
        self.assertFalse(result)

    @mock.patch('bleachbit.FileUtilities.exe_exists')
    def test_desktop_missing_wine(self, mock_exe_exists):
        """Unit test for .desktop file without Wine installed"""
        mock_exe_exists.side_effect = [
            True, False]  # First True for 'env', then False for 'wine'
        fake_config = FakeConfig(
            {"Desktop Entry": {"Exec": "env WINEPREFIX=/some/path wine notepad.exe"}})
        result = _is_broken_xdg_desktop_application(fake_config, "foo.desktop")
        self.assertTrue(result)
        mock_exe_exists.assert_has_calls([mock.call('env'), mock.call('wine')])

    @mock.patch('os.path.exists')
    @mock.patch('bleachbit.FileUtilities.exe_exists')
    def test_desktop_wine_valid_windows_exe(self, mock_exe_exists, mock_path_exists):
        """Unit test for .desktop file pointing to valid Windows application"""
        fake_config = FakeConfig({"Desktop Entry": {
                                 "Exec": r"/usr/bin/env WINEPREFIX=/some/path wine C:\\Windows\\notepad.exe"}})
        mock_exe_exists.return_value = True  # for env and wine
        mock_path_exists.return_value = True  # for notepad
        result = _is_broken_xdg_desktop_application(fake_config, "foo.desktop")
        self.assertFalse(result)

    @mock.patch('os.path.exists')
    @mock.patch('bleachbit.FileUtilities.exe_exists')
    def test_desktop_env_missing_windows_exe(self, mock_exe_exists, mock_path_exists):
        """Unit test for .desktop file with env pointing to missing Windows application"""
        fake_config = FakeConfig({"Desktop Entry": {
                                 "Exec": "env WINEPREFIX=/some/path wine_exe does_not_exist.exe"}})
        mock_exe_exists.return_value = True  # for env and wine
        mock_path_exists.return_value = False  # for does_not_exist.exe
        result = _is_broken_xdg_desktop_application(fake_config, "foo.desktop")
        self.assertTrue(result)

    def test_desktop_missing_keys(self):
        """Unit test for .desktop file missing keys"""

        test_cases = [
            ("", "Blank file"),
            ("[Desktop Entry]\n", "missing Type and Name"),
            ("[Desktop Entry]\nType=Application\n", "missing Name key"),
            ("[Desktop Entry]\nName=Test\n", "missing Type key"),
            ("[Desktop Entry]\nType=Link\nName=Test\n",
             "Type=Link and missing URL key"),
            ("[Desktop Entry]\nType=Application\nName=Test",
             "Type=Application and missing Exec"),
            ("garbage", "Not a .desktop file"),
            ("#/bin/bash\necho hello world", "Bash script")
        ]

        for content, description in test_cases:
            with tempfile.NamedTemporaryFile(mode='w+', delete=False) as tf:
                tf.write(content)
                tf.flush()
                result = is_broken_xdg_desktop(tf.name)
                self.assertTrue(result, f"Failed case: {description}")
            os.unlink(tf.name)

    @common.skipIfWindows
    def test_journald_clean(self):
        if not exe_exists('journalctl'):
            self.assertRaises(RuntimeError, journald_clean)
        else:
            try:
                journald_clean()
            except RuntimeError as rte:
                # On my system as a regular user, this succeeds.
                # On Travis running Trusty, this worked.
                # On Travis running Xenial, there is a permissions error.
                if common.have_root():
                    raise rte

    def test_journald_regex(self):
        """Test the regex for journald_clean()"""
        positive_cases = ('Vacuuming done, freed 0B of archived journals on disk.',
                          'Vacuuming done, freed 1K of archived journals on disk.',
                          'Vacuuming done, freed 100.0M of archived journals on disk.',
                          'Vacuuming done, freed 1G of archived journals on disk.',
                          'Vacuuming done, freed 0B of archived journals from /run/log/journal.',
                          'Vacuuming done, freed 1.0G of archived journals from /var/log/journal/123abc.')
        regex = re.compile(JOURNALD_REGEX)
        for pos in positive_cases:
            self.assertTrue(regex.match(pos))
        negative_cases = ('Deleted archived journal /var/log/journal/123/system@123-123.journal~ (56.0M).',
                          'Archived and active journals take up 100.0M on disk.')
        for neg in negative_cases:
            self.assertFalse(regex.match(neg))

    def test_get_purgeable_locales(self):
        """Unit test for method get_purgeable_locales()"""
        # 'en' implies 'en_US'
        locales_to_keep = ['en', 'en_AU', 'en_CA', 'en_GB']
        purgeable_locales = get_purgeable_locales(locales_to_keep)
        self.assertIsInstance(purgeable_locales, frozenset)
        self.assertIn('es', purgeable_locales)
        self.assertIn('pt', purgeable_locales)
        self.assertIn('de', purgeable_locales)
        for keep in locales_to_keep + ['en_US']:
            self.assertNotIn(keep, purgeable_locales)

        # 'en_US' keeps 'en' but discards 'en_AU'.
        purgeable_locales = get_purgeable_locales(['en_US'])
        self.assertNotIn('en', purgeable_locales)
        self.assertNotIn('en_US', purgeable_locales)
        self.assertIn('en_AU', purgeable_locales)

    def test_locale_regex(self):
        """Unit test for locale_to_language()"""
        tests = {'en': 'en',
                 'en_US': 'en',
                 'en_US@piglatin': 'en',
                 'en_US.utf8': 'en',
                 'ko_KR.eucKR': 'ko',
                 'pl.ISO8859-2': 'pl',
                 'zh_TW.Big5': 'zh'}
        regex = re.compile('^' + Locales.localepattern + '$')
        for locale, tlc in tests.items():
            m = regex.match(locale)
            self.assertIsNotNone(m, 'expected positive match for ' + locale)
            self.assertEqual(m.group("locale"), tlc)
        for test in ['default', 'C', 'English', 'ru_RU.txt', 'ru.txt']:
            self.assertIsNone(regex.match(
                test), 'expected negative match for ' + test)

    @common.skipIfWindows
    def test_localization_paths(self):
        """Unit test for localization_paths()"""
        configpath = parseString(
            '<path location="/usr/share/locale/" filter="*" />').firstChild
        self.locales.add_xml(configpath)
        counter = 0
        for path in self.locales.localization_paths(['en', 'en_AU', 'en_CA', 'en_GB']):
            self.assertLExists(path)
            self.assertTrue(path.startswith('/usr/share/locale'))
            # /usr/share/locale/en_* should be ignored
            self.assertEqual(path.find('/en_'), -1, 'expected path ' + path +
                             ' to not contain /en_')
            counter += 1
        self.assertGreater(counter, 0, 'Zero files deleted by localization cleaner. ' +
                                       'This may be an error unless you really deleted all the files.')

    @common.skipIfWindows
    def test_fakelocalizationdirs(self):
        """Create a faked localization hierarchy and clean it afterwards"""

        keepdirs = [
            'important_dontdelete',
            'important_dontdelete/ru',
            'delete',
            'delete/locale',
            'delete/locale/en',
            'delete/dummyfiles',
            'foobar',
            'foobar/locale']
        nukedirs = [
            'delete/locale/ru',
            'foobar/locale/ru']
        keepfiles = [
            'delete/dummyfiles/dontdeleteme_ru.txt',
            'important_dontdelete/exceptthisone_ru.txt',
            'delete/dummyfiles/en.txt',
            'delete/dummyfiles/ru.dic']
        nukefiles = [
            'delete/dummyfiles/ru.txt',
            'delete/locale/ru_RU.UTF8.txt']
        for path in keepdirs + nukedirs:
            os.mkdir(os.path.join(self.tempdir, path))
        for path in keepfiles + nukefiles:
            self.write_file(path)

        configxml = '<path directoryregex="^.*$">' \
                    '  <path directoryregex="^(locale|dummyfiles)$">' \
                    '    <path location="." filter="*" />' \
                    r'    <regexfilter postfix="\.txt" />' \
                    '  </path>' \
                    '</path>'
        config = parseString(configxml)
        self.locales._paths = LocaleCleanerPath(self.tempdir)
        self.locales.add_xml(config.firstChild, None)
        # normpath because paths may contain ./
        deletelist = [os.path.normpath(
            path) for path in self.locales.localization_paths(['en', 'de'])]
        for path in keepdirs + keepfiles:
            self.assertNotIn(os.path.join(self.tempdir, path), deletelist)
        for path in nukedirs + nukefiles:
            self.assertIn(os.path.join(self.tempdir, path), deletelist)

    @common.skipIfWindows
    def test_rotated_logs_real(self):
        """Unit test for rotated_logs()"""
        for path in rotated_logs():
            self.assertLExists(
                path, f"Rotated log path '{path}' does not exist")

    @mock.patch('bleachbit.FileUtilities.whitelisted')
    @mock.patch('bleachbit.FileUtilities.children_in_directory')
    def test_rotated_logs_mock(self, mock_cid, mock_whitelisted):
        mock_whitelisted.side_effect = lambda path: path.startswith(
            '/var/log/whitelisted/')
        expected_delete = [
            '/var/log/apt/history.log.1.gz',
            '/var/log/auth.log.3.old',
            '/var/log/dmesg.0',
            '/var/log/dmesg.1.gz',
            '/var/log/foo.gz',
            '/var/log/foo.old',
            '/var/log/foo/bar.0',
            '/var/log/foo/bar.gz',
            '/var/log/foo/bar.old',
            '/var/log/kern.log-20230601',
            '/var/log/messages-20090118.bz2',
            '/var/log/messages-20090118.gz',
            '/var/log/messages-20090118.xz'
            '/var/log/messages-20090118',
            '/var/log/messages.2.bz2',
            '/var/log/samba/log.smbd-20250126.gz',
            '/var/log/syslog.1',
            '/var/log/syslog.2.xz',
            '/var/log/cups/access_log.9'

        ]
        expected_keep = [
            '/var/log/dmesg',
            '/var/log/packages/foo.0',
            '/var/log/removed_packages/foo.0',
            '/var/log/removed_scripts/foo.0',
            '/var/log/samba/log.192.168.0.1',
            '/var/log/samba/log.172.17.0.1',
            '/var/log/scripts/foo.0',
            '/var/log/syslog',
            '/var/log/sysstat/sar24',
            '/var/log/whitelisted/foo.0'
        ]
        mock_cid.return_value = iter(expected_delete + expected_keep)
        result = list(rotated_logs())
        self.assertEqual(set(result), set(expected_delete))
        for path in expected_keep:
            self.assertNotIn(path, result)
        mock_cid.assert_called_once_with('/var/log')
        mock_whitelisted.assert_called()

    @common.skipIfWindows
    def test_run_cleaner_cmd(self):
        """Unit test for run_cleaner_cmd()"""
        self.assertRaises(RuntimeError, run_cleaner_cmd,
                          '/hopethisdoesntexist', [])
        # Use absolute path in case like `env -i`.
        which_sh = 'sh' if exe_exists('sh') else '/usr/bin/sh'
        self.assertRaises(subprocess.CalledProcessError, run_cleaner_cmd,
                          which_sh, ['-c', 'echo errormsg; false'])
        # test if regexes for invalid lines work
        self.assertRaises(RuntimeError, run_cleaner_cmd, 'echo', ['This is an invalid line'],
                          error_line_regexes=['invalid'])

        freed_space_regex = r'^Freed ([\d.]+[kMT]?B)'
        lines = ['Test line',
                 'Freed 100B on your hard drive',
                 'Freed 1.9kB, hooray!',
                 'Fred 12MB']
        freed_space = run_cleaner_cmd(
            'echo', ['\n'.join(lines)], freed_space_regex)
        self.assertEqual(freed_space, 2000)

    @common.skipIfWindows
    def test_wine_to_linux_path(self):
        """Unit test for wine_to_linux_path()"""
        wineprefix = "/home/foo/.wine"
        windows_pathname = "C:\\Program Files\\NSIS\\NSIS.exe"
        result = "/home/foo/.wine/drive_c/Program Files/NSIS/NSIS.exe"
        self.assertEqual(wine_to_linux_path(
            wineprefix, windows_pathname), result)

    @common.skipIfWindows
    def test_yum_clean(self):
        """Unit test for yum_clean()"""
        if 0 != os.geteuid() or os.path.exists('/var/run/yum.pid') \
                or not exe_exists('yum'):
            self.assertRaises(RuntimeError, yum_clean)
        else:
            bytes_freed = yum_clean()
            self.assertIsInteger(bytes_freed)
            logger.debug('yum bytes cleaned %d', bytes_freed)

    @common.skipIfWindows
    def test_dnf_clean(self):
        """Unit test for dnf_clean()"""
        if 0 != os.geteuid() or os.path.exists('/var/run/dnf.pid') \
                or not exe_exists('dnf'):
            self.assertRaises(RuntimeError, dnf_clean)
        else:
            bytes_freed = dnf_clean()
            self.assertIsInteger(bytes_freed)
            logger.debug('dnf bytes cleaned %d', bytes_freed)

    @common.skipIfWindows
    def test_dnf_autoremove_real(self):
        """Unit test for dnf_autoremove() with real dnf"""
        if 0 != os.geteuid() or os.path.exists('/var/run/dnf.pid') \
                or not exe_exists('dnf'):
            self.assertRaises(RuntimeError, dnf_clean)
        else:
            bytes_freed = dnf_autoremove()
            self.assertIsInteger(bytes_freed)
            logger.debug('dnf bytes cleaned %d', bytes_freed)

    @common.skipIfWindows
    @mock.patch('bleachbit.Language.setup_translation')
    @mock.patch('bleachbit.Unix.os.path')
    @mock.patch('bleachbit.General.run_external')
    def test_dnf_autoremove_mock(self, mock_run, mock_path, mock_setup):
        """Unit test for dnf_autoremove() with mock"""
        # Don't call setup_translation() for real because it uses
        # os.path.exists(), which is mocked here.
        mock_setup.return_value = None
        mock_path.exists.return_value = True
        self.assertRaises(RuntimeError, dnf_autoremove)

        mock_path.exists.return_value = False
        mock_run.return_value = (1, 'stdout', 'stderr')
        self.assertRaises(RuntimeError, dnf_autoremove)

        mock_run.return_value = (0, 'Nothing to do.', 'stderr')
        bytes_freed = dnf_autoremove()
        self.assertEqual(bytes_freed, 0)

        mock_run.return_value = (
            0, 'Remove  112 Packages\nFreed space: 299 M\n', 'stderr')
        bytes_freed = dnf_autoremove()
        self.assertEqual(bytes_freed, 299000000)

    @common.skipIfWindows
    def test_pacman_cache(self):
        """Unit test for pacman_cache()"""
        if 0 != os.geteuid() or os.path.exists('/var/lib/pacman/db.lck') \
                or not exe_exists('paccache'):
            self.assertRaises(RuntimeError, pacman_cache)
        else:
            bytes_freed = pacman_cache()
            self.assertIsInteger(bytes_freed)
            logger.debug('pacman bytes cleaned %d', bytes_freed)

    @common.skipIfWindows
    @mock.patch('bleachbit.Unix.General.run_external')
    @mock.patch('bleachbit.Unix.exe_exists')
    @mock.patch('bleachbit.Unix.os.path')
    def test_pacman_cache_mock(self, mock_path, mock_exe_exists, mock_run):
        """Unit test pacman_cache() parse logic without pacman installed"""
        mock_path.exists.return_value = False
        mock_exe_exists.return_value = True
        mock_run.return_value = (
            0,
            "==> finished: 3 packages removed (42.31 M freed)\n",
            ''
        )

        bytes_freed = pacman_cache()
        self.assertEqual(bytes_freed, 42310000)

        mock_run.assert_called_once_with(['paccache', '-rk0'])

    def test_snapd_is_active_no_snap(self):
        """Unit test for snapd_is_active() when snap is not installed"""
        with mock.patch('bleachbit.Unix.exe_exists', return_value=False):
            self.assertFalse(snapd_is_active())

    def test_snapd_is_active_no_systemctl(self):
        """Unit test for snapd_is_active() when systemctl is not installed"""
        with mock.patch('bleachbit.Unix.exe_exists', side_effect=lambda x: x == 'snap'):
            self.assertFalse(snapd_is_active())

    def test_snapd_is_active_inactive(self):
        """Unit test for snapd_is_active() when snapd.socket is inactive"""
        with mock.patch('bleachbit.Unix.exe_exists', return_value=True), \
                mock.patch('bleachbit.General.run_external', return_value=(1, '', '')):
            self.assertFalse(snapd_is_active())

    def test_snapd_is_active_active(self):
        """Unit test for snapd_is_active() when snapd.socket is active"""
        with mock.patch('bleachbit.Unix.exe_exists', return_value=True), \
                mock.patch('bleachbit.General.run_external', return_value=(0, '', '')):
            self.assertTrue(snapd_is_active())

    def test_snapd_is_active_timeout(self):
        """Unit test for snapd_is_active() timeout handling"""
        with mock.patch('bleachbit.Unix.exe_exists', return_value=True), \
                mock.patch('bleachbit.General.run_external') as mock_run:
            mock_run.side_effect = subprocess.TimeoutExpired(
                ['systemctl', 'is-active', '--quiet', 'snapd.socket'], 5)
            self.assertFalse(snapd_is_active())

    @common.skipIfWindows
    @common.skipUnlessDestructive
    def test_snap_disabled_clean(self):
        """Unit test for snap_disabled_clean()"""
        if not exe_exists('snap'):
            self.assertRaises(RuntimeError, snap_disabled_clean)
        else:
            bytes_freed = snap_disabled_clean()
            self.assertIsInteger(bytes_freed)
            logger.debug('snap disabled bytes freed %d', bytes_freed)

    @common.skipIfWindows
    def test_snap_disabled_preview(self):
        """Unit test for snap_disabled_preview()"""
        if not exe_exists('snap'):
            self.assertRaises(RuntimeError, snap_disabled_preview)
        else:
            bytes_freed = snap_disabled_preview()
            self.assertIsInstance(bytes_freed, int)
            logger.debug('snap disabled bytes freed %d', bytes_freed)

    @common.skipIfWindows
    def test_snap_timeout(self):
        """Unit test for snap_disabled_full() timeout handling"""
        with mock.patch('bleachbit.Unix.snapd_is_active', return_value=True), \
                mock.patch('bleachbit.General.run_external') as mock_run:
            # Simulate snap list --all hanging
            mock_run.side_effect = subprocess.TimeoutExpired(
                ['snap', 'list', '--all'], 10)
            with self.assertRaises(RuntimeError) as cm:
                snap_disabled_preview()
            self.assertIn('timed out', str(cm.exception))

    def test_snap_parse_list_real_data(self):
        """Unit test for snap_parse_list() with real 'snap list --all' output"""
        sample = (
            "Name                       Version                         Rev    Tracking            Publisher             Notes\n"
            "astral-uv                  0.8.9                           902    latest/stable       lengau                classic\n"
            "astral-uv                  0.7.21                          759    latest/stable       lengau                disabled,classic\n"
            "bare                       1.0                             5      latest/stable       canonical**           base\n"
            "cheese                     44.1                            78     latest/stable       ken-vandine*          -\n"
            "cheese                     44.1                            76     latest/stable       ken-vandine*          disabled\n"
        )
        expected = [
            ('astral-uv', '759'),
            ('cheese', '76')
        ]
        result = snap_parse_list(sample)
        self.assertEqual(result, expected)

    def test_snap_parse_list_no_snaps_message(self):
        """Unit test for snap_parse_list() with 'no snaps installed' output"""
        sample = "No snaps are installed yet. Try 'snap install hello-world'.\n"
        result = snap_parse_list(sample)
        self.assertEqual(result, [])

    @common.skipIfWindows
    def test_is_unix_display_protocol_wayland_no_mock(self):
        """Unit test for test_is_unix_display_protocol_wayland() with no mock"""
        ret = is_unix_display_protocol_wayland()
        self.assertIsInstance(ret, bool)

    @common.skipIfWindows
    def test_is_unix_display_protocol_wayland_mock_env(self):
        """Unit test for test_is_unix_display_protocol_wayland() by mocking os.environ"""
        # The environment variables take precedence.
        with mock.patch.dict('os.environ', {'XDG_SESSION_TYPE': 'x11'}, clear=True):
            self.assertFalse(is_unix_display_protocol_wayland())
        with mock.patch.dict('os.environ', {'XDG_SESSION_TYPE': 'wayland'}, clear=True):
            self.assertTrue(is_unix_display_protocol_wayland())
        with mock.patch.dict('os.environ', {'WAYLAND_DISPLAY': 'yes'}, clear=True):
            self.assertTrue(is_unix_display_protocol_wayland())

    @common.skipIfWindows
    @mock.patch.dict(os.environ, {}, clear=True)
    def test_is_unix_display_protocol_wayland_mock_no_env(self):
        """Unit test for test_is_unix_display_protocol_wayland() by clearing os.environ"""
        self.assertEqual(len(os.environ), 0)

        # Check that it does not throw an exception.
        is_unix_display_protocol_wayland()

        display_protocol = None

        def side_effect_func(value):
            if len(value) == 1:
                return (0, 'SESSION  UID USER   SEAT  TTY \n      2 1000 debian seat0 tty2\n\n1 sessions listed.\n', '')
            if len(value) > 1:
                return (0, f'Type={display_protocol}\n', '')
            assert False  # should never reach here

        for display_protocol, assert_method in [['wayland', self.assertTrue], ['donotexist', self.assertFalse]]:
            with mock.patch('bleachbit.General.run_external') as mock_run_external:
                mock_run_external.side_effect = side_effect_func
                is_wayland = is_unix_display_protocol_wayland()
                assert mock_run_external.called
            self.assertIsInstance(is_wayland, bool)
            assert_method(is_wayland)

        with mock.patch('bleachbit.General.run_external') as mock_run_external:
            mock_run_external.side_effect = FileNotFoundError('not found')
            self.assertFalse(is_unix_display_protocol_wayland())
            assert mock_run_external.called

        with mock.patch('bleachbit.General.run_external') as mock_run_external:
            mock_run_external.side_effect = lambda a: (1, None, None)
            self.assertFalse(is_unix_display_protocol_wayland())
            assert mock_run_external.called

        with mock.patch('bleachbit.General.run_external') as mock_run_external:
            mock_run_external.side_effect = lambda a: (
                0, 'test of invalid output from loginctl', None)
            self.assertFalse(is_unix_display_protocol_wayland())
            assert mock_run_external.called

    @common.skipIfWindows
    def test_xhost_autorized_root(self):
        error_code = None

        def side_effect_func(*args, **_kwargs):
            self.assertEqual(args[0][0], 'xhost')
            return (error_code,
                    '',
                    '')

        for error_code, expected_root_autorized in [[1, True], [0, False]]:
            with mock.patch('bleachbit.General.run_external') as mock_run_external:
                mock_run_external.side_effect = side_effect_func
                root_autorized = root_is_not_allowed_to_X_session()
            self.assertIsInstance(root_autorized, bool)
            self.assertEqual(expected_root_autorized, root_autorized)

    @common.skipIfWindows
    def test_get_trash_paths_snap_symlink(self):
        """Do not follow symlinks in ~/snap when finding trash"""
        with common.set_temporary_env('HOME', self.tempdir):
            # Create snap structure: snap/app/238/.local/share/Trash/files/
            real_rev = os.path.join(self.tempdir, 'snap', 'app', '238')
            trash_files = os.path.join(
                real_rev, '.local', 'share', 'Trash', 'files')
            os.makedirs(trash_files)
            test_file = os.path.join(trash_files, 'test.txt')
            with open(test_file, 'w') as f:
                f.write('test')
            # Create symlink current -> 238
            current_dir = os.path.join(self.tempdir, 'snap', 'app', 'current')
            os.symlink('238', current_dir)

            paths = [cmd.path for cmd in get_trash_paths()]
            # Should find the test file through the real path
            self.assertIn(test_file, paths)
            # Should NOT find it through the symlink
            symlink_path = os.path.join(
                self.tempdir, 'snap', 'app', 'current', '.local', 'share',
                'Trash', 'files', 'test.txt')
            self.assertNotIn(symlink_path, paths)


class LocalizationsTestCase(common.BleachbitTestCase):

    """Test case for localizations in Unix module"""

    @staticmethod
    def _get_recognized_paths(vfs):
        """Load localizations.xml and return paths recognized by it using the given VFS."""
        xml_path = os.path.join(os.path.dirname(
            os.path.dirname(__file__)), 'cleaners', 'localizations.xml')
        with open(xml_path, 'r', encoding='utf-8') as f:
            xml_doc = parseString(f.read())

        localizations_node = None
        for child in xml_doc.firstChild.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.nodeName == 'localizations':
                localizations_node = child
                break
        assert localizations_node is not None, 'localizations.xml must contain a <localizations> element'

        locales = Locales(vfs=vfs)
        for child in localizations_node.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                locales.add_xml(child)

        recognized = set()
        for (locale, specifier, path) in locales._paths.get_localizations('/'):
            recognized.add(path)
        return recognized

    @staticmethod
    def _path_matches_recognized(test_path, recognized):
        """Return True if test_path is in recognized or is a child/descendant of one."""
        if test_path in recognized:
            return True
        for rec_path in recognized:
            if test_path.startswith(rec_path + '/'):
                return True
        return False

    def test_localization_paths_positive(self):
        """Verify each path in localization_paths_positive.txt is matched by localizations.xml

        Uses a virtual filesystem so the test verifies actual XML rule matching
        without requiring a real /var/lib/flatpak tree.
        """
        data_file = os.path.join(os.path.dirname(
            __file__), 'localization_paths_positive.txt')
        self.assertExists(data_file)

        with open(data_file, 'r', encoding='utf-8') as f:
            test_paths = [line.strip() for line in f if line.strip()
                          and not line.strip().startswith('#')]
        self.assertGreater(len(test_paths), 0,
                           'Test data file should contain at least one path')

        vfs = ListVFS(test_paths)
        recognized = self._get_recognized_paths(vfs)

        for test_path in test_paths:
            self.assertTrue(
                self._path_matches_recognized(test_path, recognized),
                f'Path not matched by localizations.xml: {test_path}'
            )

    def test_localization_paths_negative(self):
        """Verify non-localization paths in flatpak are NOT matched

        Uses a virtual filesystem so the test verifies that paths
        that look similar to localizations are correctly rejected.
        """
        data_file = os.path.join(os.path.dirname(
            __file__), 'localization_paths_negative.txt')
        self.assertExists(data_file)

        with open(data_file, 'r', encoding='utf-8') as f:
            negative_paths = [line.strip() for line in f if line.strip(
            ) and not line.strip().startswith('#')]
        self.assertGreater(len(negative_paths), 0,
                           'Test data file should contain at least one path')

        positive_file = os.path.join(os.path.dirname(
            __file__), 'localization_paths_positive.txt')
        with open(positive_file, 'r', encoding='utf-8') as f:
            positive_paths = [line.strip() for line in f if line.strip(
            ) and not line.strip().startswith('#')]

        vfs = ListVFS(positive_paths + negative_paths)
        recognized = self._get_recognized_paths(vfs)

        for neg_path in negative_paths:
            self.assertFalse(
                self._path_matches_recognized(neg_path, recognized),
                f'Path should NOT be matched by localizations.xml: {neg_path}'
            )



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module CLI
"""

# standard imports
import copy
import datetime
import io
import locale
import os
import random
import tempfile
from unittest.mock import MagicMock, patch

# first party imports
import bleachbit
from bleachbit.CLI import (
    CliCallback,
    args_to_operations,
    args_to_operations_list,
    cleaners_list,
    parse_cmd_line,
    preview_or_clean,
    process_cmd_line)
from bleachbit.General import get_executable, run_external
from bleachbit.GtkShim import HAVE_GTK
from bleachbit import FileUtilities, Options, IS_WINDOWS, IS_POSIX
from tests import common

RUN_EXTERNAL_TIMEOUT = 30


class CLITestCase(common.BleachbitTestCase):
    """Test case for module CLI"""

    def setUp(self):
        super(CLITestCase, self).setUp()
        Options.options.reset_overrides()
        Options.options.set_override("first_start", False)

    def _test_preview(self, args, redirect_stdout=True, env=None):
        """Helper to test preview"""
        # Use devnull because in some cases the buffer will be too large,
        # and the other alternative, the screen, is not desirable.
        with open(os.devnull, 'w') as stdout:
            if not redirect_stdout:
                stdout = None
            output = run_external(args, stdout=stdout,
                                  env=env, timeout=RUN_EXTERNAL_TIMEOUT)
        self.assertEqual(output[0], 0, "Return code = %d, stderr='%s'"
                         % (output[0], output[2]))
        pos = output[2].find('Traceback (most recent call last)')
        if pos > -1:
            print("Saw the following error when using args '%s':\n %s" %
                  (args, output[2]))
        self.assertEqual(
            pos, -1, f"Traceback found in stderr when using args {args}:\n{output[2][pos:]}")

    def test_args_to_operations_list(self):
        """Unit test for args_to_operations_list()"""
        # --preset
        Options.init_configuration()
        o = args_to_operations_list(True, False)
        self.assertEqual(o, [])
        Options.options.set_tree('system', 'tmp', True)
        o = args_to_operations_list(True, False)
        self.assertEqual(o, ['system.tmp'])

        # --all-but-warning
        o = args_to_operations_list(False, True)
        self.assertIsInstance(o, list)
        self.assertTrue('google_chrome.cache' in o)
        self.assertTrue('system.tmp' in o)
        gui_available = IS_WINDOWS or HAVE_GTK
        self.assertEqual(gui_available, 'system.clipboard' in o)
        self.assertFalse('system.empty_space' in o)
        self.assertFalse('system.memory' in o)

    def test_args_to_operations(self):
        """Unit test for args_to_operations()"""
        # test explicit cleaners (without --preset or --all-but-warning)
        tests = (
            (['adobe_reader.*'], [],
             {'adobe_reader': ['cache', 'mru', 'tmp']}),
            (['adobe_reader.mru'], [], {'adobe_reader': ['mru']}),
            (['adobe_reader.*'], ['adobe_reader.cache'],
             {'adobe_reader': ['mru', 'tmp']}))
        for test_args, excludes, expected in tests:
            o = args_to_operations(test_args, False, False, excludes)
            self.assertIsInstance(o, dict)
            self.assertEqual(o, expected)

        # Test failure on wildcard
        with self.assertRaises(SystemExit) as cm:
            args_to_operations(
                ['adobe_reader.mru'], False, False, ['adobe_reader.*'])
        self.assertEqual(cm.exception.code, 1)

    def test_parse_cmd_line_except(self):
        """Unit test for parse_cmd_line() --except handling"""
        tests = (
            (['--clean', 'firefox.*', 'chromium.*',
              '--except', 'firefox.passwords,chromium.passwords'],
             ['firefox.*', 'chromium.*'],
             ['firefox.passwords', 'chromium.passwords']),
            (['--clean', 'firefox.*', 'chromium.*',
              '--except', 'firefox.passwords',
              '--except', 'chromium.passwords'],
             ['firefox.*', 'chromium.*'],
             ['firefox.passwords', 'chromium.passwords']),
            (['--clean', 'chromium.*',
              '--except', 'firefox.passwords',
              'firefox.*',
              '--except', 'chromium.passwords'],
             ['chromium.*', 'firefox.*'],
             ['firefox.passwords', 'chromium.passwords']))
        for argv, expected_args, expected_excludes in tests:
            _parser, options, args, excludes = parse_cmd_line(argv)
            self.assertTrue(options.clean)
            self.assertEqual(expected_args, args)
            self.assertEqual(expected_excludes, excludes)

    def test_cleaners_list(self):
        """Unit test for cleaners_list()"""
        for cleaner in cleaners_list():
            self.assertIsString(cleaner)

    def test_debug_log(self):
        """Unit test for --debug-log option"""

        # These texts are required in the log file.
        file_required_texts = [
            'DEBUG - Debug log file initialized',
            'ERROR - Failed to clean',
            'KeyError',
            'doesnot'
        ]

        # These texts are forbidden in stderr.
        stderror_forbidden_texts = [
            'bleachbit.CLI',
            'bleachbit.CleanerML',
            'ERROR',
            'DEBUG',
            datetime.datetime.now().strftime('%Y-%m-%d')
        ]

        for delimiter in (' ', '='):
            # delete_on_close=False is helpful but requires Python 3.12.
            with tempfile.NamedTemporaryFile(delete=False, dir=self.tempdir) as f:
                f.close()
                log_path = f.name

                # 'bleachbit.py' is used instead of '-m bleachbit.CLI'
                args = [get_executable(), 'bleachbit.py', '--preview', 'doesnot.exist',
                        '--debug'] + f'--debug-log{delimiter}{log_path}'.split(delimiter)
                (rc, _stdout, stderr) = run_external(
                    args, stdout=None, timeout=RUN_EXTERNAL_TIMEOUT)
                self.assertEqual(0, rc, f"rc={rc}, stderr={stderr}")
                self.assertExists(log_path)
                with open(log_path, 'r', encoding='utf-8') as log_file:
                    log_content = log_file.read()

                self.assertEqual(1, stderr.count('Traceback'))
                self.assertEqual(1, log_content.count(
                    'Traceback'), log_content)

                for stderr_forbidden_text in stderror_forbidden_texts:
                    self.assertNotIn(stderr_forbidden_text, stderr,
                                     f"The text '{stderr_forbidden_text}' was found in stderr")

                for required_text in file_required_texts:
                    self.assertIn(required_text, log_content,
                                  f"The text '{required_text}' was not found in the log file")
            os.remove(log_path)
            self.assertNotExists(log_path)

    @common.skipIfWindows
    def test_encoding(self):
        """Unit test for encoding"""

        filename = self.write_file(
            '/tmp/bleachbit-test-cli-encoding-\xe4\xf6\xfc~')
        # not assertExists because it doesn't cope with invalid encodings
        self.assertTrue(os.path.exists(filename))

        env = copy.deepcopy(os.environ)
        env['LANG'] = 'en_US'  # not UTF-8
        args = [get_executable(), '-m', 'bleachbit.CLI', '-p', 'system.tmp']
        # If Python pipes stdout to file or devnull, the test may give
        # a false negative.  It must print stdout to terminal.
        self._test_preview(args, redirect_stdout=False, env=env)

        os.remove(filename)
        self.assertNotExists(filename)

    def test_invalid_locale(self):
        """Unit test for invalid locales"""
        original_locale = locale.getlocale(locale.LC_NUMERIC)
        old_lang = common.get_env('LANG')
        with common.set_temporary_env('LANG', 'blahfoo'):
            # tests are run from the parent directory
            args = [get_executable(), '-m', 'bleachbit.CLI', '--version']
            output = run_external(args, timeout=RUN_EXTERNAL_TIMEOUT)
            self.assertNotEqual(output[1].find('Copyright'), -1, str(output))
        self.assertEqual(common.get_env('LANG'), old_lang)
        self.assertEqual(locale.getlocale(locale.LC_NUMERIC), original_locale)

    def test_preview(self):
        """Unit test for --preview option"""
        env = copy.deepcopy(os.environ)
        # the language and encoding affect the test results
        env['LANG'] = 'en_US.UTF-8'
        env['PYTHONIOENCODING'] = 'utf-8:backslashreplace'
        args_list = []
        module = 'bleachbit.CLI'
        big_args = [get_executable(), '-m', module, '--preview', ]
        # The full list can take a long time and generally does not improve the testing,
        # so test a subset.
        full_cleaners_list = list(cleaners_list())
        system_cleaners = [
            c for c in full_cleaners_list if c.startswith('system.')]
        if 'system.clipboard' in system_cleaners:
            # Fix error on GitHub Actions
            # FIXME: do conditional removal
            system_cleaners.remove('system.clipboard')
        non_system_cleaners = [
            c for c in full_cleaners_list
            # deepscan.* can be too slow
            if not c.startswith('system.') and not c.startswith('deepscan.')]
        sample_cleaners = random.sample(non_system_cleaners, 5)
        for cleaner in (system_cleaners + sample_cleaners):
            args_list.append(
                [get_executable(), '-m', module, '--preview', cleaner])
            big_args.append(cleaner)
        args_list.append(big_args)
        for args in args_list:
            self._test_preview(args, env=env)

    def test_delete(self):
        """Unit test for --delete option"""
        prefixes = [
            'bleachbit-test-cli-delete',
            '\x8b\x8b-bad-encoding'
        ]
        for prefix in prefixes:
            with self.subTest(prefix=prefix):
                filename = self.mkstemp(prefix=prefix)
                filename = os.path.normcase(filename)

                deleted_paths = []
                crash = [False]

                def dummy_delete(path, shred=False, crash=crash,
                                   deleted_paths=deleted_paths):
                    try:
                        self.assertLExists(path)
                    except AssertionError:
                        # On busy systems, other temp files may disappear
                        # between scan and delete, so flag a crash only for the
                        # test file itself.
                        if os.path.normcase(path) == os.path.normcase(filename):
                            crash[0] = True
                    deleted_paths.append(os.path.normcase(path))

                with patch.object(FileUtilities, 'delete', side_effect=dummy_delete):
                    FileUtilities.delete(filename)
                    # File exists because delete() is mocked, so real
                    # delete() was not called.
                    self.assertExists(filename)
                    operations = args_to_operations(['system.tmp'], False, False)
                    preview_or_clean(operations, True, quiet=True)

                self.assertIn(filename, deleted_paths,
                              f"{filename} not found deleted")
                os.remove(filename)
                self.assertNotExists(filename)
                self.assertFalse(crash[0], "Crash detected during deletion")

    def test_append_text(self):
        """Unit test for CliCallback.append_text() with special strings"""
        cb = CliCallback(quiet=False)
        for test_str in common.SPECIAL_TEST_STRINGS:
            # Test that append_text handles special strings without crashing
            cb.append_text(test_str + "\n")
            # Test with newlines stripped (as implementation does)
            cb.append_text(f"prefix{test_str}suffix\n")
            # Test tag parameter (ignored but should not crash)
            cb.append_text(test_str + "\n", _tag="test_tag")

    def test_return_text(self):
        """Check for correct text in output"""
        # format of tests is: args, expected_output, clean_env, env

        tests = [
            [['--help'], 'Usage: ', False, None],
            [['--version'], 'BleachBit version', True, None],
            [['--sysinfo'], 'sys.version', False, None]
        ]
        if IS_POSIX:
            # GUI is not available with clean environment on POSIX.
            tests.append([['--help'], 'Usage: ', True, None])
            # Force detection of Wayland
            wayland_env = os.environ.copy()
            wayland_env['XDG_SESSION_TYPE'] = 'wayland'
            tests.append([['--sysinfo'], 'sys.version', False, wayland_env])

        for args, expected_output, clean_env, env in tests:
            launcher_output = {}
            for i, launcher in enumerate([['bleachbit.py'], ['-m', 'bleachbit.CLI']]):
                direct_args = [get_executable(),] + launcher + args
                output = run_external(
                    direct_args, env=env, clean_env=clean_env, timeout=RUN_EXTERNAL_TIMEOUT)
                self.assertEqual(output[0], 0, output)
                self.assertIn(expected_output, output[1])
                modified_output_str = output[1].replace(
                    'CLI.py', 'bleachbit.py')
                launcher_output[i] = [line for line in modified_output_str.split(
                    '\n') if not line.startswith("sys.argv")]
            self.assertEqual(launcher_output[0], launcher_output[1])

    def test_process_cmd_line_mutually_exclusive(self):
        """Unit test for process_cmd_line() with mutually exclusive commands"""
        with patch('sys.argv', ['bleachbit', '--preview', '--clean']):
            with self.assertRaises(SystemExit) as cm:
                process_cmd_line()
            self.assertEqual(cm.exception.code, 1)

    def test_process_cmd_line_version(self):
        """Unit test for process_cmd_line() with --version"""
        with patch('sys.argv', ['bleachbit', '--version']):
            with patch('sys.stdout', new_callable=io.StringIO) as mock_stdout:
                with self.assertRaises(SystemExit) as cm:
                    process_cmd_line()
                self.assertEqual(cm.exception.code, 0)
            self.assertIn('BleachBit version', mock_stdout.getvalue())

    def test_process_cmd_line_list_cleaners(self):
        """Unit test for process_cmd_line() with --list-cleaners"""
        with patch('bleachbit.CLI.list_cleaners') as mock_list:
            with patch('sys.argv', ['bleachbit', '--list-cleaners']):
                with self.assertRaises(SystemExit) as cm:
                    process_cmd_line()
                self.assertEqual(cm.exception.code, 0)
            mock_list.assert_called_once()

    def test_process_cmd_line_pot(self):
        """Unit test for process_cmd_line() with --pot"""
        with patch('bleachbit.CleanerML.create_pot') as mock_create_pot:
            with patch('sys.argv', ['bleachbit', '--pot']):
                with self.assertRaises(SystemExit) as cm:
                    process_cmd_line()
                self.assertEqual(cm.exception.code, 0)
            mock_create_pot.assert_called_once()

    def test_process_cmd_line_wipe_empty_space_no_args(self):
        """Unit test for process_cmd_line() --wipe-empty-space with no args"""
        with patch('sys.argv', ['bleachbit', '--wipe-empty-space']):
            with self.assertRaises(SystemExit) as cm:
                process_cmd_line()
            self.assertEqual(cm.exception.code, 1)

    def test_process_cmd_line_preview_no_operations(self):
        """Unit test for process_cmd_line() --preview with no operations"""
        with patch('sys.argv', ['bleachbit', '--preview']):
            with self.assertRaises(SystemExit) as cm:
                process_cmd_line()
            self.assertEqual(cm.exception.code, 1)

    def test_process_cmd_line_overwrite_warning(self):
        """Unit test for process_cmd_line() --overwrite warning with --preview"""
        with patch('bleachbit.CLI.preview_or_clean'):
            with patch('bleachbit.CLI.logger.warning') as mock_warning:
                with patch('sys.argv', ['bleachbit', '--preview', 'system.tmp', '--overwrite']):
                    with self.assertRaises(SystemExit) as cm:
                        process_cmd_line()
                    self.assertEqual(cm.exception.code, 0)
                mock_warning.assert_called_once()

    def test_process_cmd_line_overwrite_no_warning_with_clean(self):
        """Unit test for process_cmd_line() --overwrite without warning when used with --clean"""
        with patch('bleachbit.CLI.preview_or_clean'):
            with patch('bleachbit.CLI.logger.warning') as mock_warning:
                with patch('sys.argv', ['bleachbit', '--clean', 'system.tmp', '--overwrite']):
                    with self.assertRaises(SystemExit) as cm:
                        process_cmd_line()
                    self.assertEqual(cm.exception.code, 0)
                mock_warning.assert_not_called()

    def test_process_cmd_line_sysinfo(self):
        """Unit test for process_cmd_line() with --sysinfo"""
        with patch('bleachbit.CLI.SystemInformation.get_system_information',
                   return_value='test sysinfo'):
            with patch('sys.stdout', new_callable=io.StringIO) as mock_stdout:
                with patch('sys.argv', ['bleachbit', '--sysinfo']):
                    with self.assertRaises(SystemExit) as cm:
                        process_cmd_line()
                    self.assertEqual(cm.exception.code, 0)
            self.assertIn('test sysinfo', mock_stdout.getvalue())

    def test_process_cmd_line_shred(self):
        """Unit test for process_cmd_line() with --shred"""
        with patch.dict('bleachbit.CLI.backends', clear=True):
            with patch('bleachbit.CLI.create_simple_cleaner', return_value=MagicMock()):
                with patch('bleachbit.CLI.preview_or_clean') as mock_preview:
                    with patch('sys.argv', ['bleachbit', '--shred', '/tmp/test-shred']):
                        with self.assertRaises(SystemExit) as cm:
                            process_cmd_line()
                        self.assertEqual(cm.exception.code, 0)
                    mock_preview.assert_called_once()

    def test_process_cmd_line_gui(self):
        """Unit test for process_cmd_line() with --gui"""
        mock_gui_module = MagicMock()
        mock_app = MagicMock()
        mock_app.run.return_value = 0
        mock_gui_module.Bleachbit.return_value = mock_app
        with patch.dict('sys.modules', {'bleachbit.GuiApplication': mock_gui_module}):
            with patch.object(bleachbit, 'GuiApplication', mock_gui_module, create=True):
                with patch('bleachbit.Bootstrap.check_wayland_and_root', return_value=False):
                    with patch('os.name', 'posix'):
                        with patch('sys.argv', ['bleachbit', '--gui']):
                            with self.assertRaises(SystemExit) as cm:
                                process_cmd_line()
                            self.assertEqual(cm.exception.code, 0)
        mock_gui_module.Bleachbit.assert_called_once_with(
            uac=False, shred_paths=[], auto_exit=None)
        mock_app.run.assert_called_once()

    def test_process_cmd_line_no_command(self):
        """Unit test for process_cmd_line() with no command"""
        # Reminder: When GUI is available and there are no arguments, then
        # process_cmd_line is not called.
        # process_cmd_line() is called when either GUI is not available or
        # there are arguments.
        with patch('sys.argv', ['bleachbit']):
            with patch('sys.stdout', new_callable=io.StringIO) as mock_stdout:
                process_cmd_line()
            self.assertIn('bleachbit', mock_stdout.getvalue())

    def test_process_cmd_line_debug(self):
        """Unit test for process_cmd_line() with --debug"""
        with patch('bleachbit.CLI.preview_or_clean'):
            with patch('sys.argv', ['bleachbit', '--debug', '--preview', 'system.tmp']):
                with self.assertRaises(SystemExit) as cm:
                    process_cmd_line()
                self.assertEqual(cm.exception.code, 0)
            self.assertTrue(Options.options.has_override('debug'))

    def test_process_cmd_line_preset(self):
        """Unit test for process_cmd_line() with --preset"""
        # Preset branch only calls set_root_log_level when debug is enabled.
        Options.options.set_override('debug', True)
        with patch('bleachbit.CLI.preview_or_clean'):
            with patch('bleachbit.CLI.set_root_log_level') as mock_set_level:
                with patch('sys.argv', ['bleachbit', '--preview', '--preset', 'system.tmp']):
                    with self.assertRaises(SystemExit) as cm:
                        process_cmd_line()
                    self.assertEqual(cm.exception.code, 0)
                mock_set_level.assert_called_once()

    def test_process_cmd_line_debug_log(self):
        """Unit test for process_cmd_line() with --debug-log"""
        with patch('bleachbit.CLI.preview_or_clean'):
            with patch('bleachbit.CLI.SystemInformation.get_system_information',
                       return_value='test sysinfo'):
                with patch('bleachbit.CLI.logger.info') as mock_log_info:
                    with patch('sys.argv', ['bleachbit', '--debug-log', '/tmp/test.log',
                                            '--preview', 'system.tmp']):
                        with self.assertRaises(SystemExit) as cm:
                            process_cmd_line()
                        self.assertEqual(cm.exception.code, 0)
                    mock_log_info.assert_any_call('test sysinfo')

    def test_process_cmd_line_option_overrides(self):
        """Unit test for process_cmd_line() option overrides"""
        with patch('sys.argv', ['bleachbit', '--no-load-cleaners',
                                '--no-delete-confirmation']):
            with patch('sys.stdout', new_callable=io.StringIO):
                process_cmd_line()
            self.assertFalse(Options.options.get('load_cleaners'))
            self.assertFalse(Options.options.get('delete_confirmation'))

    def test_shred(self):
        """Unit test for --shred"""
        suffixes = ['', '.', '.txt']
        dirs = ['.', self.tempdir]
        for dir_ in dirs:
            for suffix in suffixes:
                (fd, filename) = tempfile.mkstemp(
                    prefix='bleachbit-test-cli-shred', suffix=suffix, dir=dir_)
                os.close(fd)
                if '.' == dir_:
                    filename = os.path.basename(filename)
                # not assertExists because something strange happens on Windows
                self.assertTrue(os.path.exists(filename))
                args = [get_executable(), '-m',
                        'bleachbit.CLI', '--shred', filename]
                output = run_external(args, timeout=RUN_EXTERNAL_TIMEOUT)
                self.assertEqual(output[0], 0)
                self.assertNotExists(filename)

    @common.also_with_sudo
    def test_sysinfo(self):
        """Unit test for --sysinfo

        By not using `-m bleachbit.CLI`, this exercises `./bleachbit.py`.
        """
        env_configs = [[]]
        if not IS_WINDOWS:
            env_configs.extend([
                ['env', '-i'],  # No environment variables
                ['env', '-i', 'XDG_SESSION_TYPE=wayland']  # Mimic Wayland
            ])
        for env_prefix in env_configs:
            args = env_prefix + [get_executable(), 'bleachbit.py', '--sysinfo']
            output = run_external(args, timeout=RUN_EXTERNAL_TIMEOUT)
            if IS_POSIX and os.environ.get('USER') == 'root' and \
                    output[0] == 1:
                continue
            self.assertEqual(output[0], 0, output)
            self.assertIn('sys.version', output[1])
            # FIXME: verify that there is not a message like
            # (bleachbit.py:1234): Gdk-CRITICAL **: 23:05:08.581: gdk_screen_get_root_window: assertion 'GDK_IS_SCREEN (screen)' failed

    @common.skipUnlessWindows
    def test_gui_exit(self):
        """Unit test for --gui --exit, only for Windows"""
        args = (get_executable(), '-m',
                'bleachbit.CLI', '--gui', '--exit')
        (rc, _stdout, stderr) = run_external(
            args, timeout=RUN_EXTERNAL_TIMEOUT)
        self.assertNotIn('no such option', stderr)
        self.assertNotIn('Usage: CLI.py', stderr)
        self.assertEqual(rc, 0)
        # Is the application still running?
        opened_windows_titles = common.get_opened_windows_titles()
        self.assertFalse('BleachBit' in opened_windows_titles)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Test case for module Options
"""

import errno
import os
import stat
import subprocess
import sys
from unittest import mock

from tests import common
import bleachbit.Options
from bleachbit import IS_WINDOWS
from bleachbit.Options import _option_index
from bleachbit.Log import is_debugging_enabled_via_cli


class FakeTimer:
    """Fake timer to replace threading.Timer"""
    instances = []

    def __init__(self, interval, function, args=(), kwargs=None):
        """Init the fake timer"""
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs or {}
        self.daemon = False
        self.started = False
        self.cancelled = False
        self.__class__.instances.append(self)

    def cancel(self):
        """Cancel the fake timer."""
        self.cancelled = True

    def fire(self):
        """Fire the fake timer."""
        if not self.cancelled:
            self.function(*self.args, **self.kwargs)

    def start(self):
        """Start the fake timer."""
        self.started = True


class OptionsTestCase(common.BleachbitTestCase):
    """Test case for class Options"""

    def _write_seed_options_file(self):
        """Write a seed options file for testing"""
        with open(bleachbit.options_file, 'w', encoding='utf-8-sig') as handle:
            handle.write(f'''[bleachbit]
version={bleachbit.APP_VERSION}
shred=False
check_online_updates=True
[hashpath]
[list/shred_drives]
''')

    def test_option_index(self):
        """Unit test for _option_index"""
        self.assertEqual(_option_index('0'), 0)
        self.assertEqual(_option_index('3'), 3)
        self.assertEqual(_option_index('1_type'), 1)
        self.assertEqual(_option_index('2_path'), 2)

    def test_Options(self):
        """Unit test for class Options"""
        o = bleachbit.Options.options
        value = o.get("check_online_updates")

        # toggle a boolean
        o.toggle('check_online_updates')
        self.assertEqual(not value, o.get("check_online_updates"))

        # restore original boolean
        o.set("check_online_updates", value)
        self.assertEqual(value, o.get("check_online_updates"))

        # test delayed commit
        shred = o.get("shred")
        o.set("shred", not shred)
        self.assertEqual(not shred, o.get("shred"))

        o.restore()  # abandon uncommitted changes
        self.assertEqual(shred, o.get("shred"))

        o.set("shred", shred)
        o.close()  # commit
        self.assertEqual(shred, o.get("shred"))

        # try a list
        list_values = ['a', 'b', 'c']
        o.set_list("list_test", list_values)
        self.assertEqual(list_values, o.get_list("list_test"))

        # these should always be set
        for bkey in bleachbit.Options.boolean_keys:
            self.assertIsInstance(o.get(bkey), bool)

        # language
        value = o.get_language('en')
        self.assertIsInstance(value, bool)
        o.set_language('en', True)
        self.assertTrue(o.get_language('en'))
        o.set_language('en', False)
        self.assertFalse(o.get_language('en'))
        o.set_language('en', value)

        # tree
        o.set_tree("parent", "child", True)
        self.assertTrue(o.get_tree("parent", "child"))
        o.set_tree("parent", "child", False)
        self.assertFalse(o.get_tree("parent", "child"))
        o.config.remove_option("tree", "parent.child")
        self.assertFalse(o.get_tree("parent", "child"))

    def test_keep_list(self):
        """Test for get_whitelist_paths() / set_whitelist_paths()"""
        o = bleachbit.Options.options
        # FIXME: finish renaming whitelist to keep list
        self.assertIsInstance(o.get_whitelist_paths(), list)
        keep_list = [('file', '/home/foo'), ('folder', '/home'),
                     ('file', '/home/unicode/кодирование')]
        old_keep_list = o.get_whitelist_paths()
        o.config.remove_section('whitelist/paths')
        self.assertIsInstance(o.get_whitelist_paths(), list)
        self.assertEqual(o.get_whitelist_paths(), [])
        o.set_whitelist_paths(keep_list)
        self.assertIsInstance(o.get_whitelist_paths(), list)
        self.assertEqual(set(keep_list), set(o.get_whitelist_paths()))
        o.set_whitelist_paths(old_keep_list)
        o.close()
        self.assertEqual(set(old_keep_list), set(o.get_whitelist_paths()))

    def test_init_configuration(self):
        """Test for init_configuration()"""
        if os.path.exists(bleachbit.options_file):
            os.remove(bleachbit.options_file)
        self.assertNotExists(bleachbit.options_file)
        bleachbit.Options.init_configuration()
        self.assertExists(bleachbit.options_file)

    def test_is_corrupt(self):
        """Test is_corrupt()"""
        def _test_is_corrupt(contents, expect_is_corrupt):
            with open(bleachbit.options_file, 'w', encoding='utf-8-sig') as f:
                f.write(contents)
            o = bleachbit.Options.Options()
            self.assertEqual(o.is_corrupt(), expect_is_corrupt)

        # test blank
        _test_is_corrupt('', False)
        _test_is_corrupt('[bleachbit]\n', False)

        # test valid but non-standard boolean
        _test_is_corrupt('[bleachbit]\nshred=f\n', False)

        # test invalid boolean
        # https://github.com/bleachbit/bleachbit/issues/560#issuecomment-497361700
        _test_is_corrupt("[bleachbit]\nshred=['True']\n", True)
        os.remove(bleachbit.options_file)

    def test_purge(self):
        """Test purging"""
        # By default ConfigParser stores keys (the filenames) as lowercase.
        # This needs special consideration when combined with purging.
        o1 = bleachbit.Options.Options()
        pathname = self.write_file('foo.xml')
        myhash = '0ABCD'
        o1.set_hashpath(pathname, myhash)
        o1.close()
        self.assertEqual(myhash, o1.get_hashpath(pathname))
        if IS_WINDOWS:
            # check case sensitivity
            self.assertEqual(myhash, o1.get_hashpath(pathname.upper()))

        # reopen
        o2 = bleachbit.Options.Options()
        # write something, which triggers the purge
        o2.set('dummypath', 'dummyvalue', 'hashpath')
        o2.commit()
        # verify the path was not purged
        self.assertTrue(os.path.exists(pathname))
        self.assertEqual(myhash, o2.get_hashpath(pathname))

        # delete the path
        os.remove(pathname)
        # close and reopen
        o2.close()
        o3 = bleachbit.Options.Options()
        # write something, which triggers the purge
        o3.set('dummypath', 'dummyvalue', 'hashpath')
        o3.commit()
        # verify the path was purged
        self.assertIsNone(o3.get_hashpath(pathname))
        o3.close()

    def test_abbreviations(self):
        """Test non-standard, abbreviated booleans T and F"""

        # set values
        o = bleachbit.Options.options
        if not o.config.has_section('test'):
            o.config.add_section('test')
        o.config.set('test', 'test_t_upper', 'T')
        o.config.set('test', 'test_f_upper', 'F')
        o.config.set('test', 'test_t_lower', 't')
        o.config.set('test', 'test_f_lower', 'f')

        # read
        self.assertEqual(o.config.getboolean('test', 'test_t_upper'), True)
        self.assertEqual(o.config.getboolean('test', 'test_t_lower'), True)
        self.assertEqual(o.config.getboolean('test', 'test_f_upper'), False)
        self.assertEqual(o.config.getboolean('test', 'test_f_lower'), False)

        # clean up
        del o

    def test_percent(self):
        """Test that the percent sign can be used without quoting the string"""
        # https://github.com/bleachbit/bleachbit/issues/205

        opt = bleachbit.Options.options.config
        if not opt.has_section('test'):
            opt.add_section('test')
        opt.set('test', 'filename', '/var/log/samba/log.%m')

        # read
        self.assertEqual(opt.get('test', 'filename'), '/var/log/samba/log.%m')

        # clean up
        del opt

    def test_error_disk_full(self):
        """Test graceful degradation when disk is full"""
        disk_full_error = OSError('No space left on device')
        disk_full_error.errno = errno.ENOSPC
        o = bleachbit.Options.Options()
        with mock.patch('builtins.open', side_effect=disk_full_error):
            with self.assertLogs(level='ERROR') as log_context:
                o.set('test_key', 'test_value')
                o.commit()
        self.assertIn('Disk was full', log_context.output[0])
        self.assertIn(bleachbit.options_file, log_context.output[0])
        self.assertEqual(o.get('test_key'), 'test_value')

    def test_error_permission(self):
        """Test graceful degradation with permission errors"""
        permission_error = PermissionError('Permission denied')
        permission_error.errno = errno.EACCES
        o = bleachbit.Options.Options()
        with mock.patch('builtins.open', side_effect=permission_error):
            with self.assertLogs(level='ERROR') as log_context:
                o.set('test_key', 'test_value')
                o.commit()
        self.assertIn('Permission denied', log_context.output[0])
        self.assertIn(bleachbit.options_file, log_context.output[0])
        self.assertEqual(o.get('test_key'), 'test_value')

    def test_warning(self):
        """Test warning preferences"""
        o = bleachbit.Options.options
        self.assertFalse(o.get_warning_preference('test_key'))
        o.remember_warning_preference('test_key')
        self.assertTrue(o.get_warning_preference('test_key'))
        o.close()

    def test_warning_keys_with_colons(self):
        """Test restoring warning preferences with colons in keys"""
        with open(bleachbit.options_file, 'w', encoding='utf-8-sig') as handle:
            handle.write('''[bleachbit]
[warnings]
cleaner:google_chrome:passwords = True
cleaner:microsoft_edge:passwords = True
''')
        o = None
        try:
            if sys.version_info >= (3, 10):
                with self.assertNoLogs('bleachbit.Options', level='ERROR'):
                    o = bleachbit.Options.Options()
            else:
                o = bleachbit.Options.Options()
            self.assertTrue(o.get_warning_preference(
                'cleaner:google_chrome:passwords'))
            self.assertTrue(o.get_warning_preference(
                'cleaner:microsoft_edge:passwords'))
        finally:
            if o is not None:
                o.close()

    def test_corrupt_warning_preferences_are_migrated(self):
        """Test migration of warning preferences corrupted by colon parsing"""
        with open(bleachbit.options_file, 'w', encoding='utf-8-sig') as handle:
            handle.write('''[bleachbit]
[warnings]
cleaner = google_chrome:passwords = True
protected_path = /tmp = True
''')
        o = bleachbit.Options.Options()
        try:
            self.assertTrue(o.get_warning_preference(
                'cleaner:google_chrome:passwords'))
            self.assertTrue(o.get_warning_preference('protected_path:/tmp'))
            self.assertFalse(o.config.has_option('warnings', 'cleaner'))
            self.assertFalse(o.config.has_option('warnings', 'protected_path'))
            o.commit()
        finally:
            o.close()
        with open(bleachbit.options_file, 'r', encoding='utf-8-sig') as handle:
            contents = handle.read()
        self.assertIn('cleaner:google_chrome:passwords = True', contents)
        self.assertIn('protected_path:/tmp = True', contents)
        self.assertNotIn('cleaner = google_chrome:passwords = True', contents)
        self.assertNotIn('protected_path = /tmp = True', contents)

    def test_overrides(self):
        """Test CLI override functionality"""
        o = bleachbit.Options.options

        debug_default = is_debugging_enabled_via_cli()

        # Check defaults which are assumed in later tests.
        self.assertFalse(o.get('shred'))
        self.assertEqual(o.get('debug'), debug_default)
        self.assertTrue(o.get('delete_confirmation'))

        # Test basic override
        o.set_override('shred', True)
        self.assertTrue(o.get('shred'))

        # Test that override takes precedence over get
        o.set('shred', False)
        self.assertTrue(o.get('shred'))  # override should still return True

        # Test multiple overrides
        o.set_override('debug', True)
        o.set_override('delete_confirmation', False)
        self.assertTrue(o.get('debug'))
        self.assertFalse(o.get('delete_confirmation'))

        # Test that flush doesn't write overridden values
        # First set persistent values different from overrides
        o.set('shred', False)
        o.set('debug', False)
        o.set('delete_confirmation', True)
        o.commit()

        # Verify overrides still work after flush
        self.assertTrue(o.get('shred'))
        self.assertTrue(o.get('debug'))
        self.assertFalse(o.get('delete_confirmation'))

        # Create new instance to verify persistent values were written correctly
        o2 = bleachbit.Options.Options()
        self.assertFalse(o2.get('shred'))
        self.assertEqual(o2.get('debug'), debug_default)
        self.assertTrue(o2.get('delete_confirmation'))
        o2.close()

    def test_override_does_not_persist(self):
        """Test that overrides don't persist after process restart"""
        o = bleachbit.Options.options

        # Set an override
        o.set_override('shred', True)
        self.assertTrue(o.get('shred'))

        # Create new instance (simulating process restart)
        o2 = bleachbit.Options.Options()
        # Override should not exist in new instance
        self.assertNotIn(('bleachbit', 'shred'), o2.overrides)
        o2.close()

        # Clean up
        o.reset_overrides()

    def test_override_with_sections(self):
        """Test overrides work with different sections"""
        o = bleachbit.Options.options

        # Test override in non-default section
        if not o.config.has_section('test_section'):
            o.config.add_section('test_section')
        o.config.set('test_section', 'test_key', 'original_value')

        o.set_override('test_key', 'override_value', section='test_section')
        self.assertEqual(
            o.get('test_key', section='test_section'), 'override_value')

        # Verify flush doesn't write override
        o.commit()
        o2 = bleachbit.Options.Options()
        self.assertEqual(
            o2.get('test_key', section='test_section'), 'original_value')
        o2.close()

        # Clean up
        o.reset_overrides()
        if o.config.has_section('test_section'):
            o.config.remove_section('test_section')

    def test_flush_mechanisms_timer_and_close(self):
        """Test that changes flush correctly via timer fire or explicit close()."""
        with mock.patch('bleachbit.Options.threading.Timer', FakeTimer):
            for commit_method in ['timer', 'close']:
                with self.subTest(commit=commit_method):
                    self._write_seed_options_file()
                    o = bleachbit.Options.Options()

                    def assert_pending_changes_uncommitted(instance):
                        self.assertFalse(instance.get('shred'))
                        self.assertIsNone(instance.get_list('list_test'))
                        self.assertEqual([], instance.get_whitelist_paths())
                        self.assertEqual([], instance.get_custom_paths())
                        self.assertFalse(instance.get_language('zz'))
                        self.assertFalse(instance.get_tree('system', 'cache'))
                        self.assertFalse(
                            instance.get_warning_preference('warn_key'))

                    # Ensure initial state has no pending changes
                    assert_pending_changes_uncommitted(o)

                    # Guard against premature flush
                    with mock.patch('builtins.open', side_effect=AssertionError('unexpected flush')):
                        o.set('shred', True)
                        o.set_list('list_test', ['a', 'b'])
                        o.set_whitelist_paths([('file', '/tmp/a')])
                        o.set_custom_paths([('folder', '/tmp/b')])
                        o.set_language('zz', True)
                        o.set_tree('system', 'cache', True)
                        o.remember_warning_preference('warn_key')

                    pending_timer = FakeTimer.instances[-1]
                    self.assertTrue(pending_timer.started)

                    # Verify uncommitted changes are invisible to new instance
                    o2 = bleachbit.Options.Options()
                    assert_pending_changes_uncommitted(o2)
                    o2.close()

                    # Commit via the specific mechanism being tested
                    if commit_method == 'timer':
                        pending_timer.fire()
                    else:
                        o.close()

                    # Verify all changes persisted
                    o3 = bleachbit.Options.Options()
                    self.assertTrue(o3.get('shred'))
                    self.assertEqual(['a', 'b'], o3.get_list('list_test'))
                    self.assertEqual([('file', '/tmp/a')],
                                     o3.get_whitelist_paths())
                    self.assertEqual([('folder', '/tmp/b')],
                                     o3.get_custom_paths())
                    self.assertTrue(o3.get_language('zz'))
                    self.assertTrue(o3.get_tree('system', 'cache'))
                    self.assertTrue(o3.get_warning_preference('warn_key'))
                    o3.close()

                    if commit_method == 'timer':
                        o.close()

                    # Clean up for next iteration
                    self.tearDown()
                    self.setUp()

    def test_close_unregisters_atexit_and_is_idempotent(self):
        """Test close() unregisters its atexit callback and only flushes once."""
        self._write_seed_options_file()
        with mock.patch('bleachbit.Options.atexit.unregister') as mock_unregister:
            o = bleachbit.Options.Options()
            with mock.patch.object(o, '_Options__flush') as mock_flush:
                o.close()
                o.close()

            mock_flush.assert_called_once_with(force=False)
            mock_unregister.assert_called_once()

    def test_unicode_paths(self):
        """Test that paths with various Unicode characters"""
        # On Linux, os.listdir() may return filenames with surrogate-escaped
        # bytes (e.g. invalid UTF-8 byte sequences decoded with
        # 'surrogateescape').  These contain unpaired surrogates (U+D800 to
        # U+DFFF) which cannot be encoded as UTF-8, causing UnicodeEncodeError
        # when ConfigParser writes the .ini file.
        # https://github.com/bleachbit/bleachbit/issues/2082

        test_cases = [
            ('surrogate', '/tmp/invalid_unicode_surrogate_\udce9'),
            ('cjk', '/tmp/中文日本語한국어'),
            ('european', '/tmp/àéïöüñçß'),
            ('cyrillic', '/tmp/кодирование'),
            ('arabic', '/tmp/العربية'),
            ('emoji', '/tmp/🎉🚀💻'),
            ('mixed', '/tmp/mixed_中文_àéï_\udce9'),
        ]

        def verify_paths(o, keep_list, custom_list):
            """Helper to verify whitelist and custom paths"""
            o.set_whitelist_paths(keep_list)
            o.commit()
            self.assertEqual(keep_list, o.get_whitelist_paths())

            o.set_custom_paths(custom_list)
            o.commit()
            self.assertEqual(custom_list, o.get_custom_paths())

            # Verify round-trip: new instance reads from disk
            o2 = bleachbit.Options.Options()
            self.assertEqual(keep_list, o2.get_whitelist_paths())
            self.assertEqual(custom_list, o2.get_custom_paths())
            o2.close()

        o = bleachbit.Options.options
        for name, path in test_cases:
            with self.subTest(unicode_type=name):
                keep_list = [('file', path)]
                custom_list = [('folder', path)]
                verify_paths(o, keep_list, custom_list)

        o.close()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module Network
"""

# standard imports
import ipaddress
import logging
import os
import random
import unittest
from unittest.mock import MagicMock, Mock, patch

# third party imports
import requests

# first party imports
import bleachbit
from tests import common
from bleachbit import IS_WINDOWS
from bleachbit.FileUtilities import delete
from bleachbit.Network import (download_url_to_fn, fetch_url, get_gtk_version,
                               get_ip_for_url, get_user_agent, unset_sslkeylogfile)

logger = logging.getLogger(__name__)


def response_to_error_msg(response):
    """Convert response to error message"""
    return "URL: {url}\n" + \
           f"Status: {response.status_code} {response.reason}\n" + \
           f"Response headers: {dict(response.headers)}\n" + \
           f"Response content: {response.text}"


class NetworkTestCase(common.BleachbitTestCase):
    """Test case for module Network"""
    status_generators = [
        'https://browsergym.bleachbit.org/api/status/{}',
        'https://httpbin.org/status/{}',
        'https://httpbingo.org/status/{}',
        'https://mock.httpstatus.io/{}',
        'https://postman-echo.com/status/{}',
        'https://the-internet.herokuapp.com/status_codes/{}'
    ]
    status_generator_url = None

    @classmethod
    def setUpClass(cls):
        super(NetworkTestCase, cls).setUpClass()
        status_generators = list(cls.status_generators)
        random.shuffle(status_generators)
        for generator in status_generators:
            url = generator.format(200)
            try:
                response = fetch_url(url, timeout=5, max_retries=0)
                if response.status_code == 200:
                    cls.status_generator_url = generator
                    logger.info('Using status generator: %s',
                                cls.status_generator_url)
                    return
                else:
                    logger.warning('Status generator %s returned %s',
                                   generator, response.status_code)
            except requests.exceptions.RequestException as e:
                logger.warning('Status generator failed: %s (%s)',
                               generator.format('...'), e)
        if not cls.status_generator_url:
            raise RuntimeError('No working HTTP status code generator found.')

    def setUp(self):
        """Set up the test environment before each test method."""
        super().setUp()

    def test_unset_sslkeylogfile(self):
        """Test the function unset_sslkeylogfile()."""
        with common.set_temporary_env('SSLKEYLOGFILE', 'ssl.log'):
            self.assertEqual(unset_sslkeylogfile(True), IS_WINDOWS)
            self.assertFalse(unset_sslkeylogfile(True))

    def test_download_url_to_fn(self):
        """Unit test for function download_url_to_fn()"""
        # Test different HTTP status codes
        # 200: success
        # 404: not found (non-retryable error)
        # 500: server error (retryable error)
        tests = ((self.status_generator_url.format(200), True),
                 (self.status_generator_url.format(404), False),
                 (self.status_generator_url.format(500), False))
        fn = os.path.join(self.tempdir, 'download')
        on_error_called = [False]

        def on_error(msg1, msg2):
            # Only print a simplified error message to avoid excessive output
            error_type = 'HTTP status' if 'HTTP status code' in str(
                msg2) else 'Connection'
            print(
                f'test on_error: {error_type} error for {msg1.split(":")[-1].strip()}')
            on_error_called[0] = True
        for (url, expected_rc) in tests:
            with self.subTest(url=url, expected_rc=expected_rc):
                self.assertNotExists(fn)
                on_error_called = [False]
                rc = download_url_to_fn(url, fn, None,
                                        on_error=on_error, timeout=20)
                err_msg = f'test_download_url_to_fn({url}) returned {rc} instead of {expected_rc}'
                self.assertEqual(rc, expected_rc, err_msg)
                if expected_rc:
                    self.assertExists(fn)
                self.assertNotEqual(rc, on_error_called[0])
                # minimal parameters
                rc = download_url_to_fn(url, fn)
                self.assertEqual(rc, expected_rc, err_msg)
                delete(fn, ignore_missing=True)

    def test_get_gtk_version(self):
        """Unit test for get_gtk_version()"""
        gtk_ver = get_gtk_version()
        if gtk_ver is None:
            self.skipTest("GTK is not installed")
        self.assertIsInstance(gtk_ver, str)
        self.assertRegex(gtk_ver, r"^\d+\.\d+\.\d+$")

    def test_get_ip_for_url(self):
        """Unit test for get_ip_for_url()"""
        for good_url in ('https://www.example.com', bleachbit.update_check_url):
            ip_str = get_ip_for_url(good_url)
            _ = ipaddress.ip_address(ip_str)
        for bad_url in (None, '', 'https://test.invalid'):
            ret = get_ip_for_url(bad_url)
            self.assertEqual(ret[0], '(',
                             f'get_ip_for_url({bad_url})={ret}')

    def test_get_user_agent(self):
        """Unit test for method get_user_agent()"""
        agent = get_user_agent()
        logger.debug("user agent = '%s'", agent)
        self.assertIsString(agent)

    def test_fetch_url_nonretry(self):
        """Unit test for fetch_url() without retry"""
        status_codes = (200, 404)
        for status_code in status_codes:
            url = self.status_generator_url.format(status_code)
            with self.subTest(status_code=status_code):
                response = fetch_url(url, max_retries=0, timeout=5)
                error_msg = response_to_error_msg(response)
                self.assertEqual(response.status_code, status_code,
                                 error_msg)

    def test_fetch_url_headers(self):
        """Unit test for fetch_url() header handling"""
        with patch('bleachbit.Network.requests.Session') as mock_session:
            session_instance = MagicMock()
            session_instance.get.return_value = Mock()
            mock_session.return_value.__enter__.return_value = session_instance

            tests = [
                # (custom_headers, expected_custom_key)
                (None, None),
                ({'X-BleachBit-Version': '5.1.0'}, 'X-BleachBit-Version'),
            ]
            for custom_headers, expected_custom_key in tests:
                with self.subTest(custom_headers=custom_headers):
                    response = fetch_url('https://example.com', max_retries=0, timeout=5,
                                         headers=custom_headers)
                    kwargs = session_instance.get.call_args.kwargs
                    headers = kwargs['headers']
                    self.assertEqual(
                        response, session_instance.get.return_value)
                    self.assertIsInstance(headers, dict)
                    self.assertIn('User-Agent', headers)
                    self.assertEqual(headers['User-Agent'], get_user_agent())
                    self.assertNotIn('X-OS-Type', headers)
                    if expected_custom_key:
                        self.assertIn(expected_custom_key, headers)
                        self.assertEqual(headers[expected_custom_key],
                                         custom_headers[expected_custom_key])
                    else:
                        self.assertNotIn('X-BleachBit-Version', headers)

    def test_fetch_url_retry(self):
        """Unit test for fetch_url() with retry"""
        url = self.status_generator_url.format(500)
        try:
            fetch_url(url, max_retries=1, timeout=5)
        except requests.exceptions.RetryError:
            return
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout) as exc:
            self.skipTest(f"Status generator unavailable ({exc})")
        self.fail('fetch_url() did not raise RetryError for HTTP 500 response')

    def test_fetch_url_invalid(self):
        """Unit test for fetch_url() with invalid URL"""
        url = 'https://test.invalid'
        with self.assertRaises(requests.exceptions.RequestException):
            fetch_url(url)


class MissingPackagesTestCase(unittest.TestCase):
    """Test behavior when optional third-party packages are missing."""

    def test_missing_requests(self):
        """Network should be importable without requests."""
        with common.mock_missing_package(
                'requests', 'urllib3',
                clear_prefixes=('bleachbit.Network', 'bleachbit.Update')):
            import bleachbit.Network as Network
            self.assertFalse(Network.HAVE_REQUESTS)
            with self.assertRaises(Network.RequestException):
                Network.fetch_url('https://example.com')

    def test_missing_urllib3(self):
        """Missing urllib3 should set HAVE_URLLIB3 to False."""
        with common.mock_missing_package(
                'urllib3',
                clear_prefixes=('bleachbit.Network',)):
            import bleachbit.Network as Network
            self.assertFalse(Network.HAVE_URLLIB3)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.


"""
Common code for unit tests
"""

import contextlib
import os
import re
import shutil
import sys
import tempfile
import time
import unittest
import warnings
from pathlib import Path
from unittest import mock

import bleachbit

if bleachbit.IS_WINDOWS:
    import winreg
    import win32gui
    from bleachbit import Windows
import bleachbit.Options
from bleachbit.Bootstrap import bootstrap
from bleachbit.FileUtilities import (
    children_in_directory,
    extended_path,
    is_hard_link,
    is_normal_directory,
)
from bleachbit.General import gc_collect, sudo_mode

# /etc/locale.alias may list the qaa-qtz range, which is reserved for
# private use rather than a concrete locale. Skip it if present.
SKIP_ALIAS_CODES = {'qaa-qtz', 'it_CARES'}


def _supports_stdout_char(char: str) -> bool:
    """Return True if sys.stdout can encode the given character."""
    encoding = (
        getattr(bleachbit, 'stdout_encoding', None)
        or getattr(sys.stdout, 'encoding', None)
        or sys.getdefaultencoding()
    )
    try:
        char.encode(encoding)
    except (UnicodeEncodeError, LookupError):
        return False
    return True


@contextlib.contextmanager
def mock_missing_package(*package_names, clear_prefixes=()):
    """Context manager that simulates missing optional packages.

    Saves and restores sys.modules around the block.  Removes all
    modules matching any of *package_names* or *clear_prefixes* from
    sys.modules, then patches each package_name as None to prevent
    re-import.

    Args:
        *package_names: Package names to make unavailable
            (e.g. 'requests', 'psutil').  These are both evicted from
            sys.modules and patched as None.
        clear_prefixes: Additional module prefixes to evict from
            sys.modules (e.g. 'bleachbit.Network') so they are
            re-imported inside the block.  Not patched as None.
    """
    saved_modules = dict(sys.modules)
    all_prefixes = list(package_names) + list(clear_prefixes)
    for mod in list(sys.modules.keys()):
        if any(mod.startswith(p) for p in all_prefixes):
            del sys.modules[mod]
    try:
        patch_dict = {name: None for name in package_names}
        with mock.patch.dict('sys.modules', patch_dict):
            yield
    finally:
        for mod in list(sys.modules.keys()):
            if mod not in saved_modules:
                del sys.modules[mod]
        for mod, mod_obj in saved_modules.items():
            if mod not in sys.modules:
                sys.modules[mod] = mod_obj


@contextlib.contextmanager
def capture_glib_exceptions():
    """Capture exceptions swallowed by GLib virtual method/signal handlers.

    PyGObject catches exceptions raised inside GObject virtual method
    overrides (e.g. ``do_activate``) and signal callbacks, then logs them
    via ``sys.excepthook`` without propagating to the Python caller.  This
    context manager installs a custom ``sys.excepthook`` that collects such
    exceptions so they can be re-raised after GTK event processing.

    Yields a list of ``(exc_type, exc_value, exc_tb)`` tuples.
    """
    captured = []
    original_excepthook = sys.excepthook

    def _capturing_excepthook(exc_type, exc_value, exc_tb):
        captured.append((exc_type, exc_value, exc_tb))
        original_excepthook(exc_type, exc_value, exc_tb)

    sys.excepthook = _capturing_excepthook
    try:
        yield captured
    finally:
        sys.excepthook = original_excepthook


@contextlib.contextmanager
def set_temporary_env(env_var, env_value):
    """
    Temporarily overrides an environment variable.
    """
    # Save the original value so we can restore it later
    original_value = os.environ.get(env_var)

    if env_value is None:
        os.environ.pop(env_var, None)
    else:
        os.environ[env_var] = str(env_value)
    try:
        yield
    finally:
        # Restore the original state
        if original_value is None:
            # If the environment variable wasn't set originally, remove it
            os.environ.pop(env_var, None)
        else:
            os.environ[env_var] = original_value


class BleachbitTestCase(unittest.TestCase):
    """TestCase class with several convenience methods and asserts"""
    _patchers = []

    @classmethod
    def setUpClass(cls):
        """Do common setup for the test case

        * Create a temporary directory for the testcase.
        * Treat warnings as errors.
        This is also set by environment variable in `Makefile` and
        `appveyor.yml`.
        * Patch options paths.
        """
        bootstrap()
        warnings.simplefilter("error")
        cls.tempdir = tempfile.mkdtemp(prefix=cls.__name__)
        if 'BLEACHBIT_TEST_OPTIONS_DIR' not in os.environ:
            cls._patch_options_paths()
        bleachbit.Options.options.reset_overrides()
        bleachbit.Options.options.set_override("first_start", False)

    @classmethod
    def _patch_options_paths(cls):
        to_patch = [('bleachbit.options_dir', cls.tempdir),
                    ('bleachbit.options_file', os.path.join(
                        cls.tempdir, "bleachbit.ini")),
                    ('bleachbit.personal_cleaners_dir', os.path.join(cls.tempdir, "cleaners"))]
        for target, source in to_patch:
            patcher = mock.patch(target, source)
            patcher.start()
            cls._patchers.append(patcher)

        bleachbit.Options.options.restore()

    @classmethod
    def tearDownClass(cls):
        """Do common teardown for the test case

        * Collect garbage.
        * Remove the temporary directory.
        * Restore options paths.
        """
        bleachbit.Options.options.reset_overrides()
        # Cancel any pending deferred flush (Options.__schedule_flush)
        # so its background timer does not recreate bleachbit.ini inside
        # tempdir while rmtree is mid-way through deleting it.
        bleachbit.Options.options.cancel_pending_flush()
        gc_collect()
        # On Windows, a file may be temporarily locked, so retry.
        for attempt in range(5):
            try:
                if os.path.exists(cls.tempdir):
                    shutil.rmtree(cls.tempdir)
                break
            except PermissionError:
                if attempt < 4:
                    time.sleep(1)
                else:
                    raise
        if 'BLEACHBIT_TEST_OPTIONS_DIR' not in os.environ:
            cls._stop_patch_options_paths()

    @classmethod
    def _stop_patch_options_paths(cls):
        for patcher in cls._patchers:
            patcher.stop()

    def run(self, result=None):
        """Run the test case with conditional timer message"""
        start = time.perf_counter()
        outcome = super().run(result)
        duration = time.perf_counter() - start
        threshold = os.getenv('BLEACHBIT_SLOW_TEST_THRESHOLD')  # in seconds
        if threshold:
            threshold = float(threshold)
        if not threshold or threshold < 0:
            threshold = 30.0
        if duration >= threshold:
            test_id = f"{self.__class__.__name__}.{self._testMethodName}"
            prefix = "🐌 " if _supports_stdout_char("🐌") else ""
            print(f"{prefix}SLOW TEST: {test_id} ({duration:.1f}s)", flush=True)
        return outcome

    def setUp(self):
        """Call before each test method"""
        basedir = os.path.join(os.path.dirname(__file__), '..')
        os.chdir(basedir)

    #
    # type asserts
    #

    def assertIsInteger(self, obj, msg=''):
        self.assertIsInstance(obj, int, msg)

    def assertIsString(self, obj, msg=''):
        self.assertIsInstance(obj, str, msg)

    def assertIsBytes(self, obj, msg=''):
        self.assertIsInstance(obj, bytes, msg)

    def assertIsLanguageCode(self, lang_id):
        self.assertIsInstance(lang_id, str)
        if lang_id in ('C', 'C.UTF-8', 'C.utf8', 'POSIX'):
            return
        self.assertTrue(len(lang_id) >= 2)
        pattern = r'^[a-z]{2,3}([_-]([A-Z][A-Za-z]{1,3}|[0-9]{3}))?(\.[a-zA-Z][a-zA-Z0-9-]+)?(@\w+)?$'
        self.assertTrue(re.match(pattern, lang_id),
                        f'Invalid language code format: {lang_id}')

    @staticmethod
    def check_exists(func, path):
        try:
            func(path)
            return True
        except PermissionError:
            # Python 3.4: on Windows os.path.[l]exists may return False when access is denied:
            # https://bugs.python.org/issue28075
            return True
        except:
            return False

    #
    # file asserts
    #
    @staticmethod
    def _assert_path(path):
        """Normalize a path for existence checks in unit tests."""
        # TestMakefile.py uses relative path without an environment variable.
        # TestWinapp.py uses variable "$bbtestdir"
        # However, do not expand paths that are already absolute or Path objects.
        if isinstance(path, Path):
            return str(path)
        assert isinstance(
            path, str), f'path must be a string or Path, not {type(path)}'
        if not os.path.isabs(path):
            path = os.path.expandvars(path)
        return path

    # Our assertion method names follow the convention in Python's unittest
    # pylint: disable-next=invalid-name
    def assertExists(self, path, msg='', func=os.stat):
        """Check that a file, directory, or any path exists"""
        path = self._assert_path(path)
        if not self.check_exists(func, get_test_path(path)):
            raise AssertionError(
                'The file %s should exist, but it does not. %s' % (path, msg))

    # pylint: disable-next=invalid-name
    def assertNotExists(self, path, msg='', func=os.stat):
        """Check that a file, directory, or any path does not exist"""
        path = self._assert_path(path)
        if self.check_exists(func, get_test_path(path)):
            raise AssertionError(
                'The file %s should not exist, but it does. %s' % (path, msg))

    # pylint: disable-next=invalid-name
    def assertLExists(self, path, msg=''):
        """Check that a file, directory, or any path exists using lstat"""
        self.assertExists(path, msg, os.lstat)

    # pylint: disable-next=invalid-name
    def assertNotLExists(self, path, msg=''):
        """Check that a file, directory, or any path does not exist using lstat"""
        self.assertNotExists(path, msg, os.lstat)

    def assertCondExists(self, cond, path, msg=''):
        if cond:
            self.assertExists(path, msg)
        else:
            self.assertNotExists(path, msg)

    def assertDirectoryCount(self, path, count, list_directories=True):
        """Assert that a directory has a specific number of files

        - Counts recursively
        - Counts links
        - Does not recurse links

        """
        object_list = list(children_in_directory(path, list_directories))
        self.assertEqual(len(object_list), count, f"contains {len(object_list)} objects "
                         f"such as {object_list[:2]}, expected {count}")

    #
    # file creation functions
    #
    def write_file(self, filename, contents=b'', mode='wb', encoding=None, text=None):
        """Create a temporary file, optionally writing contents to it

        If `text` is given, it is written in text mode with utf-8 encoding,
        and `mode`/`encoding` are set automatically. `text` is mutually
        exclusive with `contents`.

        The temporary file is automatically deleted after testing.
        """
        if text is not None:
            if contents != b'':
                raise ValueError("write_file: `text` is exclusive to `contents`")
            contents = text
            mode = 'w'
            encoding = 'utf-8'
        if not encoding and mode == 'w':
            encoding = 'utf-8'
        if not os.path.isabs(filename):
            filename = os.path.join(self.tempdir, filename)
        with open(extended_path(filename), mode, encoding=encoding) as f:
            f.write(contents)
        assert (os.path.exists(extended_path(filename)))
        with open(extended_path(filename), 'rb' if 'b' in mode else 'r', encoding=encoding if 'b' not in mode else None) as f:
            written_contents = f.read()
        expected = contents if 'b' in mode else contents.encode(
            encoding) if encoding else contents
        actual = written_contents if 'b' in mode else written_contents.encode(
            encoding) if encoding else written_contents
        assert actual == expected, f"File contents mismatch: expected {expected!r}, got {actual!r}"
        return filename

    def mkdir(self, dirname):
        """Create a directory with a given name

        If dirname is not absolute, it will be created in self.tempdir.

        Returns the path to the created directory.
        """
        assert isinstance(dirname, str)
        if not os.path.isabs(dirname):
            dirname = os.path.join(self.tempdir, dirname)
        ext_dirname = extended_path(dirname)
        os.makedirs(ext_dirname, exist_ok=True)
        self.assertExists(ext_dirname)
        self.assertTrue(os.path.isdir(ext_dirname))
        self.assertFalse(is_hard_link(ext_dirname))
        self.assertTrue(is_normal_directory(ext_dirname))
        self.assertFalse(os.path.isfile(ext_dirname))
        if bleachbit.IS_WINDOWS:
            self.assertFalse(Windows.is_junction(ext_dirname))
        return dirname

    def mkstemp(self, **kwargs):
        """Create a temporary file

        If dir is not specified, it will be created in self.tempdir, and tempdir
        will be automatically deleted after testing.

        If prefix is not specified, it's defined by the test method name.
        """
        if 'dir' not in kwargs:
            kwargs['dir'] = self.tempdir
        if 'prefix' not in kwargs:
            kwargs['prefix'] = f"{self.__class__.__name__}-{self._testMethodName}-"
        (fd, filename) = tempfile.mkstemp(**kwargs)
        os.close(fd)
        return filename

    def mkdtemp(self, **kwargs):
        """Create a temporary directory

        If dir is not specified, it will be created in self.tempdir, and tempdir
        will be automatically deleted after testing.

        If prefix is not specified, it's defined by the test method name.
        """
        if 'dir' not in kwargs:
            kwargs['dir'] = self.tempdir
        if 'prefix' not in kwargs:
            kwargs['prefix'] = f"{self.__class__.__name__}-{self._testMethodName}-"
        return tempfile.mkdtemp(**kwargs)


def get_test_path(path):
    """Normalize test paths for Windows"""
    if not bleachbit.IS_WINDOWS:
        return path
    path = os.path.normpath(path)
    # The \\?\ extended-length prefix applied by extended_path() requires
    # an absolute path: Windows treats "\\?\<relative>" as invalid and
    # reports it as non-existent.
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    return extended_path(path)


def get_env(key):
    """Get an environment variable. If not set, returns None instead of KeyError."""
    if not key in os.environ:
        return None
    return os.environ[key]


def have_root():
    """Return true if we have root privileges on POSIX systems"""
    return sudo_mode() or os.getuid() == 0


def put_env(key, val):
    """Put an environment variable. None removes the key

    Returns None
    """
    if not val:
        if key in os.environ:
            del os.environ[key]
    else:
        os.environ[key] = val


def skipIfWindows(f):
    """Skip unit test if running on Windows"""
    return unittest.skipIf(bleachbit.IS_WINDOWS, 'running on Windows')(f)


def skipUnlessDestructive(f):
    """Skip unless destructive tests are allowed"""
    return unittest.skipUnless(os.getenv('DESTRUCTIVE_TESTS') == 'T', 'environment variable DESTRUCTIVE_TESTS not set to T')(f)


def skipUnlessWindows(f):
    """Skip unit test unless running on Windows"""
    return unittest.skipUnless(bleachbit.IS_WINDOWS, 'not running on Windows')(f)


def also_with_sudo(test_func):
    """
    Decorator to mark test methods that should be run both normally and with sudo.

    See also `tests/test_with_sudo.py`.
    """
    test_func._also_with_sudo = True
    return test_func


def touch_file(filename):
    """Create an empty file"""
    dname = os.path.dirname(filename)
    if dname and not os.path.exists(dname):
        # Make the directory, if it does not exist.
        os.makedirs(dname)
    Path(filename).touch()
    assert (os.path.exists(filename))
    assert not is_normal_directory(filename)


def validate_result(self, result, really_delete=False):
    """Validate the command returned valid results"""
    self.assertIsInstance(result, dict, "result is a %s" % type(result))
    # label
    self.assertIsString(result['label'])
    self.assertGreater(len(result['label'].strip()), 0)
    # n_*
    self.assertIsInteger(result['n_deleted'])
    self.assertGreaterEqual(result['n_deleted'], 0)
    self.assertLessEqual(result['n_deleted'], 1)
    self.assertEqual(result['n_special'] + result['n_deleted'], 1)
    # size
    self.assertIsInstance(result['size'], (int, type(
        None),), "size is %s" % str(result['size']))
    # path
    filename = result['path']
    if not filename:
        # the process action, for example, does not have a filename
        return
    self.assertIsInstance(filename, (str, type(None)),
                          "Filename is invalid: '%s' (type %s)" % (filename, type(filename)))
    if isinstance(filename, str) and not filename[0:2] == 'HK':
        if really_delete:
            self.assertNotLExists(filename)
        else:
            self.assertLExists(filename)


def get_winregistry_value(key, subkey):
    try:
        with winreg.OpenKey(key, subkey) as hkey:
            return winreg.QueryValue(hkey, None)
    except FileNotFoundError:
        return None


def get_opened_windows_titles():
    """
    Get the titles of all opened windows.

    Returns:
        list: A list of window titles.
    """
    opened_windows_titles = []

    def enumerate_opened_windows_titles(hwnd, _ctx):
        text = win32gui.GetWindowText(hwnd)
        if win32gui.IsWindowVisible(hwnd) and text:
            opened_windows_titles.append(text)

    win32gui.EnumWindows(enumerate_opened_windows_titles, None)
    return opened_windows_titles


# Common test strings for filename testing across different test modules
# https://github.com/bleachbit/bleachbit/issues/1709
SPECIAL_TEST_STRINGS = [
    '.prefixandsuffix',  # simple
    "x".zfill(150),  # long
    ' begins_with_space',
    "''",  # quotation mark
    "~`!@#$%^&()-_+=x",  # non-alphanumeric characters
    "[]{};'.,x",  # non-alphanumeric characters
    'abcdefgh',  # simple Unicode
    'J\xf8rgen Scandinavian',
    '\u2014em-dash',  # LP#1454030
    "עִבְרִית",  # Hebrew
    "アメリカ",  # Katakana
    "ÄäǞǟËëḦḧÏïḮḯÖöȪȫṎṏT̈ẗÜüǕǖǗǘǙǚǛǜṲṳṺṻẄẅẌẍŸÿ",  # umlauts
    'sigil-should$not-change',
    'issue_1709_\udcd6',  # GitHub issue 1709
    'invalid_unicode_surrogate_\udce9',
    'multi_surrogate_\udcd6\udcd7\udcd8',
    'fire_emoji_\U0001F525',
    'ascii.bak',
    'äöüßÄÖÜ.bak',
    "עִבְרִית.bak",
    'ɡælɪk.bak'
]

# Additional strings for POSIX systems.
# Windows doesn't allow or requires special handling for these characters.
POSIX_SPECIAL_TEST_STRINGS = [
    '"*',
    '\t\\',
    ':?<>|',
    ' ',
    'endspace ',
    'endperiod.'  # Windows filenames cannot end with space or period
]



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (c) 2008-2026 Andrew Ziem.
#
# This work is licensed under the terms of the GNU GPL, version 3 or
# later.  See the COPYING file in the top-level directory.

"""
Test case for module ProtectedPath
"""

import os
import tempfile
import unittest

from functools import wraps

from tests import common
from bleachbit.ProtectedPath import (
    _check_exempt,
    _expand_path,
    _get_protected_path_xml,
    _normalize_for_comparison,
    calculate_impact,
    check_protected_path,
    clear_cache,
    get_warning_message,
    load_protected_paths)
from bleachbit import ProtectedPath as protected_path_module
from bleachbit import get_share_path
from bleachbit import IS_WINDOWS, IS_POSIX
from bleachbit.Cleaner import backends
from tests.TestCleaner import register_all_cleaners


CASE_METHODS = (
    str.lower,
    str.upper,
    str.title,
    str.swapcase,
    lambda x: x)


def requirePPXML(test_func):
    """Decorator to skip test if protected path XML is not found"""
    @wraps(test_func)
    def wrapper(self):
        if not _get_protected_path_xml():
            self.skipTest('Protected path XML not found')
        return test_func(self)
    return wrapper


class ProtectedPathTestCase(common.BleachbitTestCase):
    """Test case for ProtectedPath module"""

    def setUp(self):
        """Call before each test method"""
        super().setUp()
        clear_cache()

    def tearDown(self):
        """Call after each test method"""
        super().tearDown()
        clear_cache()

    def test_expand_path(self):
        """Test _expand_path function"""
        # Test user home expansion
        home = os.path.expanduser('~')
        result = _expand_path('~')
        self.assertEqual(result, home)

        # Test with subpath
        result = _expand_path('~/Documents')
        expected = os.path.normpath(os.path.join(home, 'Documents'))
        self.assertEqual(result, expected)

        # Test environment variable expansion
        with common.set_temporary_env('TEST_PROTECTED_PATH_VAR', '/test/path'):
            if IS_WINDOWS:
                result = _expand_path('%TEST_PROTECTED_PATH_VAR%')
            else:
                result = _expand_path('$TEST_PROTECTED_PATH_VAR')
        self.assertEqual(result, os.path.normpath('/test/path'))

    def test_normalize_for_comparison(self):
        """Test _normalize_for_comparison function"""
        # Case sensitive
        result = _normalize_for_comparison('/Home/User', True)
        self.assertIn('Home', result)

        # Case insensitive
        result = _normalize_for_comparison('/Home/User', False)
        self.assertEqual(result, result.lower())

    def test_get_protected_path_xml(self):
        """Test that protected path XML file can be found"""
        xml_path = get_share_path('protected_path.xml')
        self.assertIsNotNone(
            xml_path, 'protected_path.xml not found via get_share_path')
        self.assertExists(xml_path)
        xml_path2 = _get_protected_path_xml()
        self.assertIsNotNone(xml_path2, 'Protected path XML not found')
        self.assertExists(xml_path2)
        self.assertTrue(xml_path2.endswith('protected_path.xml'))

    @requirePPXML
    def test_load_protected_paths(self):
        """Test loading protected paths from XML"""
        paths = load_protected_paths()
        self.assertIsInstance(paths, list)
        # Should have loaded some paths for current OS
        self.assertGreater(len(paths), 0)

        # Each path should have required keys
        for ppath in paths:
            self.assertIn('path', ppath)
            self.assertIn('depth', ppath)
            self.assertIn('case_sensitive', ppath)
            self.assertIsInstance(ppath['path'], str)
            self.assertTrue(isinstance(
                ppath['depth'], int) or ppath['depth'] is None)
            self.assertIsInstance(ppath['case_sensitive'], bool)

    @requirePPXML
    @common.skipIfWindows
    def test_expand_path_entries_split_os_pathsep(self):
        """Environment variables with os.pathsep values expand into multiple entries"""

        share1 = f"/nonexistent_xdg_share1_{self._testMethodName}"
        share2 = f"/nonexistent_xdg_share2_{self._testMethodName}"
        expected_paths = [os.path.normpath(share1), os.path.normpath(share2)]
        with common.set_temporary_env('XDG_DATA_DIRS', os.pathsep.join(expected_paths)):
            clear_cache()
            paths = load_protected_paths(force_reload=True)
            matches = [pp['path']
                       for pp in paths if pp['path'] in expected_paths]
            self.assertCountEqual(matches, expected_paths)

            for candidate in expected_paths:
                result = check_protected_path(candidate)
                self.assertIsNotNone(result)
        clear_cache()

    @requirePPXML
    def test_load_protected_paths_caching(self):
        """Test that protected paths are cached"""
        paths1 = load_protected_paths()
        paths2 = load_protected_paths()
        # Should be the same object (cached)
        self.assertIs(paths1, paths2)

        # Force reload should return new list
        paths3 = load_protected_paths(force_reload=True)
        self.assertIsNot(paths1, paths3)

    @common.skipUnlessWindows
    def test_check_exempt_windows(self):
        """Test check_exempt() on Windows"""
        temp_dir = _expand_path('%temp%')
        self.assertIsInstance(temp_dir, str)
        self.assertTrue(len(temp_dir) > 0)
        for case_method in CASE_METHODS:
            cased_temp_dir = case_method(temp_dir)
            self.assertTrue(_check_exempt(cased_temp_dir),
                            f"Failed for {cased_temp_dir}")
            cased_temp_dir += os.sep
            self.assertTrue(_check_exempt(cased_temp_dir),
                            f"Failed for {cased_temp_dir}")
            git_dir = os.path.join(temp_dir, '.git')
            self.assertTrue(_check_exempt(git_dir),
                            f"Failed for {git_dir}")
            git_dir += os.sep
            self.assertTrue(_check_exempt(git_dir),
                            f"Failed for {git_dir}")

    @common.skipIfWindows
    def test_check_exempt_unix(self):
        """Test check_exempt() on Unix"""
        expect_exempt = (
            '~/.cache',
            '~/.cache/',
            '~/.cache/.git',
            '/tmp',
            '/tmp/',
            '/tmp/.git')
        for path in expect_exempt:
            expanded_path = _expand_path(path)
            self.assertTrue(_check_exempt(expanded_path), expanded_path)
            self.assertFalse(_check_exempt(
                expanded_path.upper()), expanded_path)
        expect_not_exempt = (
            '/',
            '/home',
            '/home/',
            '~/.caches',
            '~/.caches/',
            '~/.caches/.git',
            '~/Documents/',
            '/tmps',
            '/tmps/',
            '/tmps/.git')
        for path in expect_not_exempt:
            self.assertFalse(_check_exempt(_expand_path(path)), path)

    def test_check_protected_path_no_match(self):
        """Test check_protected_path with non-protected path"""
        # A random temp path should not be protected
        with tempfile.TemporaryDirectory() as tmpdir:
            result = check_protected_path(tmpdir)
            self.assertIsNone(result)
        # Check .git (which is protected) under an exempt directory
        if IS_POSIX:
            exempt_dir_raw = '~/.cache'
        elif IS_WINDOWS:
            exempt_dir_raw = '%temp%'
        else:
            self.skipTest("Unsupported OS")
        exempt_dir = _expand_path(exempt_dir_raw)
        self.assertIsInstance(exempt_dir, str)
        self.assertTrue(_check_exempt(exempt_dir))
        self.assertNotEqual(exempt_dir, exempt_dir_raw)
        self.assertTrue(len(exempt_dir) > 0)
        result = check_protected_path(exempt_dir)
        self.assertIsNone(result)
        subpath = os.path.join(exempt_dir, '.git')
        result = check_protected_path(subpath)
        self.assertIsNone(result)
        # ~/Documents/Zoom/recording.mp4 is not protected
        zoom_path_raw = '~/Documents/Zoom/recording.mp4'
        zoom_path = _expand_path(zoom_path_raw)
        self.assertNotEqual(zoom_path, zoom_path_raw)
        result = check_protected_path(zoom_path)
        self.assertIsNone(result)

    @requirePPXML
    def test_check_protected_path_exact_match(self):
        """Test check_protected_path with exact match"""
        home = os.path.expanduser('~')
        home_result = check_protected_path(home)
        documents_path = os.path.expanduser('~/Documents')
        documents_result = check_protected_path(documents_path)
        for result in (home_result, documents_result):
            self.assertIsNotNone(result)
            self.assertIn('path', result)
            self.assertIn('depth', result)

    @requirePPXML
    def test_check_protected_path_child_match(self):
        """Test check_protected_path with child of protected path

        If /home/foo/.mozilla is protected with depth 1, adding /home/foo/.mozilla/firefox
        should also warn because it is a child of the protected directory.
        """
        # If home is protected with depth > 0, children should match
        paths = load_protected_paths()

        # Find a protected path with depth > 0
        for ppath in paths:
            if ppath['depth'] > 0:
                child_path = os.path.join(ppath['path'], 'test_subdir')
                result = check_protected_path(child_path)
                if result is not None:
                    self.assertGreater(result['depth'], 0)
                break

    @requirePPXML
    def test_check_protected_path_parent_match(self):
        """Test that parent of protected path triggers warning

        If /home/foo/.mozilla is protected with any depth, adding /home/foo
        should also warn because it is a parent containing the protected path.

        """
        paths = load_protected_paths()
        if not paths:
            self.skipTest("No protected paths loaded")

        # Find a protected path that has a parent directory
        for ppath in paths:
            protected = ppath['path']
            parent = os.path.dirname(protected)
            if parent and parent != protected:
                result = check_protected_path(parent)
                # Parent should match because it contains protected path
                if result is not None:
                    self.assertIsNotNone(result)
                break

    @requirePPXML
    def test_check_protected_path_git_variations(self):
        """Test that .git entries are detected at various nesting levels"""
        git_paths = [
            '/home/test/.git',
            '/home/test/bleachbit/.git',
            '/home/test/projects/bleachbit/.git',
        ]

        for path in git_paths:
            result = check_protected_path(path)
            self.assertIsNotNone(result, msg=f"Expected .git match for {path}")
            self.assertEqual(result['path'], '.git')

    @requirePPXML
    def test_check_protected_path_desktop_suffix(self):
        """Ensure Desktop.old variants do not match Desktop protected entry"""
        desktop = os.path.expanduser('~/Desktop')
        paths = [
            desktop + '.old',
            os.path.join(desktop + '.old', 'photo.jpg'),
        ]

        for path in paths:
            result = check_protected_path(path)
            self.assertIsNone(
                result, msg=f"Unexpected Desktop match for {path}")

    def test_calculate_impact_nonexistent(self):
        """Test calculate_impact with non-existent path"""
        result = calculate_impact('/nonexistent/path/12345')
        self.assertEqual(result['file_count'], 0)
        self.assertEqual(result['total_size'], 0)
        self.assertEqual(result['size_human'], '0B')

    def test_calculate_impact_file(self):
        """Test calculate_impact with a file"""
        with tempfile.NamedTemporaryFile(delete=False, dir=self.tempdir) as f:
            f.write(b'x' * 1000)
            temp_path = f.name

        try:
            result = calculate_impact(temp_path)
            self.assertEqual(result['file_count'], 1)
            self.assertGreaterEqual(result['total_size'], 1000)
            self.assertIsInstance(result['size_human'], str)
        finally:
            os.unlink(temp_path)

    def test_calculate_impact_directory(self):
        """Test calculate_impact with a directory"""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create some test files
            for i in range(5):
                filepath = os.path.join(tmpdir, f'file{i}.txt')
                with open(filepath, 'w', encoding='utf-8') as f:
                    f.write('x' * 100)

            result = calculate_impact(tmpdir)
            self.assertEqual(result['file_count'], 5)
            self.assertGreaterEqual(result['total_size'], 500)
            self.assertIsInstance(result['size_human'], str)

    def test_calculate_impact_nested_directory(self):
        """Test calculate_impact with nested directories"""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create nested structure
            subdir = os.path.join(tmpdir, 'subdir')
            os.makedirs(subdir)

            # Files in root
            with open(os.path.join(tmpdir, 'root.txt'), 'w', encoding='utf-8') as f:
                f.write('x' * 100)

            # Files in subdir
            with open(os.path.join(subdir, 'sub.txt'), 'w', encoding='utf-8') as f:
                f.write('x' * 200)

            result = calculate_impact(tmpdir)
            self.assertEqual(result['file_count'], 2)
            self.assertGreaterEqual(result['total_size'], 300)

    def test_get_warning_message(self):
        """Test get_warning_message function"""
        impact = {
            'file_count': 10,
            'total_size': 1024000,
            'size_human': '1MB',
        }

        msg = get_warning_message('/home/user', impact)
        self.assertIsInstance(msg, str)
        self.assertIn('/home/user', msg)
        self.assertIn('10', msg)  # file count
        self.assertIn('1MB', msg)  # size

    def test_get_warning_message_no_files(self):
        """Test get_warning_message with no files"""
        impact = {
            'file_count': 0,
            'total_size': 0,
            'size_human': '0B',
        }

        msg = get_warning_message('/some/path', impact)
        self.assertIsInstance(msg, str)
        self.assertIn('/some/path', msg)

    @requirePPXML
    def test_clear_cache(self):
        """Test clear_cache function"""
        # Load to populate cache
        load_protected_paths()
        self.assertIsNotNone(protected_path_module._protected_paths_cache)

        # Clear cache
        clear_cache()
        self.assertIsNone(protected_path_module._protected_paths_cache)

    @requirePPXML
    def test_preview_all(self):
        """Check whether any real cleaners return protected paths"""
        register_all_cleaners()

        def check_option(co_key, co_option_id):
            for cmd in backends[co_key].get_commands(co_option_id):
                try:
                    for ret in cmd.execute(False):  # False = preview mode
                        if isinstance(ret, dict) and ret.get('path'):
                            path = ret['path']
                            result = check_protected_path(path)
                            if result is not None:
                                print(
                                    f'warning: {path} is protected by {result}')
                                return
                except:
                    continue
        option_counter = 0
        for key in sorted(backends):
            # loop over cleaners
            for (option_id, _option_name) in backends[key].get_options():
                # loop over options
                check_option(key, option_id)
                option_counter += 1
        print(f'checked {option_counter} options')

    @requirePPXML
    def test_path_with_trailing_slash(self):
        """Test that paths with trailing slashes are handled correctly"""
        home = os.path.expanduser('~')
        home_with_slash = home + os.sep

        result1 = check_protected_path(home)
        result2 = check_protected_path(home_with_slash)
        self.assertIsNotNone(result1)
        self.assertIsNotNone(result2)
        self.assertEqual(result1['path'], result2['path'])

    @requirePPXML
    @common.skipUnlessWindows
    def test_case_insensitive_windows(self):
        """Test case-insensitive matching on Windows"""
        profile_var = r'%userprofile%'
        profile = os.path.expandvars(profile_var)
        self.assertNotEqual(profile, profile_var)
        for case_method in CASE_METHODS:
            # the whole path is case-insensitive
            cased_profile = case_method(profile)
            self.assertIsNotNone(check_protected_path(cased_profile))

    @requirePPXML
    def test_case_insensitive(self):
        """Test case insensitive cross platform"""
        home = os.path.expanduser('~')
        self.assertIsInstance(home, str)
        protected_base = os.path.join(home, 'Desktop')
        for case_method in CASE_METHODS:
            fname = 'thesis.pdf'
            check_path = os.path.join(protected_base, case_method(fname))
            result = check_protected_path(check_path)
            self.assertIsNotNone(
                result, f'Failed to find protected path for {check_path}: {result}')
            self.assertEqual(result['path'].lower(), protected_base.lower())


if __name__ == '__main__':
    unittest.main()



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Test cases for module CleanerML
"""

# standard imports
import os
import shutil
import sys
from unittest import mock

# first party imports
import bleachbit
from tests import common
from bleachbit import Cleaner
from bleachbit.CleanerML import (
    CleanerML,
    boolstr_to_bool,
    create_pot,
    default_vars,
    list_cleanerml_files,
    load_cleaners,
    pot_fragment)


class CleanerMLTestCase(common.BleachbitTestCase):
    """Test cases for CleanerML"""

    def run_all(self, xmlcleaner, really_delete):
        """Helper function to execute all options in a cleaner"""
        for (option_id, __name) in xmlcleaner.cleaner.get_options():
            for cmd in xmlcleaner.cleaner.get_commands(option_id):
                for result in cmd.execute(really_delete):
                    common.validate_result(self, result, really_delete)

    def _get_xmlcleaner(self):
        """Helper for CleanerML*()"""
        xmlcleaner = CleanerML("doc/example_cleaner.xml")
        self.assertIsInstance(xmlcleaner, CleanerML)
        self.assertIsInstance(xmlcleaner.cleaner, Cleaner.Cleaner)
        return xmlcleaner

    def test_CleanerML(self):
        """Unit test for class CleanerML"""
        xmlcleaner = self._get_xmlcleaner()
        # preview
        self.run_all(xmlcleaner, False)

    @common.skipUnlessDestructive
    def test_CleanerML_destructive(self):
        """Unit test the destructive parts of class CleanerML"""
        xmlcleaner = self._get_xmlcleaner()
        # really delete
        self.run_all(xmlcleaner, True)

    def test_boolstr_to_bool(self):
        """Unit test for boolstr_to_bool()"""
        tests = [('True', True),
                 ('False', False)]

        for (arg, output) in tests:
            self.assertEqual(boolstr_to_bool(arg), output)
            self.assertEqual(boolstr_to_bool(arg.lower()), output)
            self.assertEqual(boolstr_to_bool(arg.upper()), output)

    def test_create_pot(self):
        """Unit test for create_pot()"""
        os.chdir('po')
        try:
            create_pot()
        finally:
            os.chdir('..')

    def test_default_vars_windows_system(self):
        """Unit test WindowsSystem in default_vars()"""
        env = {
            'WinDir': r'C:\Windows',
            'ProgramFiles': r'C:\Program Files (x86)',
            'ProgramW6432': r'C:\Program Files',
        }
        with mock.patch('bleachbit.CleanerML.os.name', 'nt'), \
                mock.patch.dict(os.environ, env, clear=True), \
                mock.patch('bleachbit.Windows.ARCH_BITS', 32):
            variables = default_vars()
        self.assertEqual(
            [r'C:\Windows\Sysnative', r'C:\Windows\SysWOW64'],
            variables['WindowsSystem'])

    def test_list_cleanerml_files(self):
        """Unit test for list_cleanerml_files()"""
        for pathname in list_cleanerml_files():
            self.assertExists(pathname)

    def test_load_cleaners(self):
        """Unit test for load_cleaners()"""
        # normal
        list(load_cleaners())

        # should catch exception with invalid XML
        pcd = bleachbit.personal_cleaners_dir
        bleachbit.personal_cleaners_dir = self.mkdtemp(
            prefix='bleachbit-cleanerml-load')
        self.write_file(os.path.join(bleachbit.personal_cleaners_dir, 'invalid.xml'),
                        contents=b'<xml><broken>')
        list(load_cleaners())
        shutil.rmtree(bleachbit.personal_cleaners_dir)
        bleachbit.personal_cleaners_dir = pcd

    def test_load_cleaners_invalid_utf8(self):
        """Unit test for load_cleaners() with invalid UTF-8 encoding"""
        pcd = bleachbit.personal_cleaners_dir
        bleachbit.personal_cleaners_dir = self.mkdtemp(
            prefix='bleachbit-cleanerml-utf8')
        self.write_file(os.path.join(bleachbit.personal_cleaners_dir, 'broken_encoding.xml'),
                        contents=b'<cleaner id="poison">\n\xff\xfe\xfd Broken\n')
        list(load_cleaners())
        shutil.rmtree(bleachbit.personal_cleaners_dir)
        bleachbit.personal_cleaners_dir = pcd

    def test_nvalid_utf8(self):
        """Test CleanerML() with invalid UTF-8 encoding

        It should fail gracefully.
        """
        fn = os.path.join(self.mkdtemp(prefix='bleachbit-cleanerml-utf8'),
                          'broken.xml')
        self.write_file(fn, contents=b'<cleaner id="poison">\n\xff\xfe\xfd\n')
        xmlcleaner = CleanerML(fn)
        self.assertIsInstance(xmlcleaner, CleanerML)
        self.assertFalse(xmlcleaner.cleaner.is_usable())

    def test_utf8_non_ascii(self):
        """Test CleanerML() with UTF-8 non-ASCII text

        It should load successfully.
        """
        xml_str = """<?xml version="1.0" encoding="UTF-8"?>
<cleaner id="test_utf8">
    <label>Test</label>
    <!-- 中文注释 -->
    <option id="opt">
        <label>测试标签</label>
        <description>测试描述</description>
        <action search="file" command="delete" path="C:\\中文路径\\file.txt"/>
    </option>
</cleaner>
"""
        fn = os.path.join(self.mkdtemp(prefix='bleachbit-cleanerml-utf8'),
                          'utf8.xml')
        self.write_file(fn, contents=xml_str.encode('utf-8'))
        xmlcleaner = CleanerML(fn)
        self.assertIsInstance(xmlcleaner, CleanerML)
        self.assertTrue(xmlcleaner.cleaner.is_usable())
        self.assertEqual('测试标签', xmlcleaner.cleaner.options['opt'][0])
        self.assertEqual('测试描述',
                         xmlcleaner.cleaner.options['opt'][1])
        commands = list(xmlcleaner.cleaner.get_commands('opt'))
        self.assertEqual(0, len(commands))

    def test_os_match(self):
        """Unit test for os_match"""
        xmlcleaner = CleanerML("doc/example_cleaner.xml")

        # blank always matches
        self.assertTrue(xmlcleaner.os_match(""))

        # as Linux
        self.assertFalse(xmlcleaner.os_match('windows', 'linux'))
        self.assertTrue(xmlcleaner.os_match('linux', 'linux'))
        self.assertTrue(xmlcleaner.os_match('unix', 'linux'))

        # as Windows
        self.assertFalse(xmlcleaner.os_match('linux', 'win32'))
        self.assertFalse(xmlcleaner.os_match('unix', 'win32'))
        self.assertTrue(xmlcleaner.os_match('windows', 'win32'))

        # as unknown operating system
        with self.assertRaises(RuntimeError):
            xmlcleaner.os_match('linux', 'hal9000')

    def test_pot_fragment(self):
        """Unit test for pot_fragment()"""
        self.assertIsString(pot_fragment("Foo", 'bar.xml'))

    def test_var(self):
        """Test the <var> element"""
        xml_str = r"""
<cleaner id="testvar">
    <label>cleaner label</label>
    <description>cleaner description</description>
    <var name="basepath">
        <value>%%LocalAppData%%\FooDoesNotExist</value>
        <value>~/.config/FooDoesNotExist</value>
        <value>{tempdir}/a</value>
        <value>{tempdir}/b</value>
    </var>
    <option id="option1">
        <label>option1 label</label>
        <description>option1 description</description>
        <action search="file" command="delete" path="$$basepath$$/test.log" />
    </option>
</cleaner>
""".format(**{'tempdir': self.tempdir})
        # write XML cleaner
        cml_path = os.path.join(self.tempdir, 'test.xml')
        self.write_file(cml_path, xml_str.encode(sys.getdefaultencoding()))

        # create two canaries
        test_log_path_a = os.path.join(self.tempdir, 'a', 'test.log')
        test_log_path_b = os.path.join(self.tempdir, 'b', 'test.log')
        common.touch_file(test_log_path_a)
        common.touch_file(test_log_path_b)
        self.assertExists(test_log_path_a)
        self.assertExists(test_log_path_b)

        # parse XML to XML cleaner instance
        xmlc = CleanerML(cml_path)
        self.assertIsInstance(xmlc, CleanerML)
        self.assertIsInstance(xmlc.cleaner, Cleaner.Cleaner)
        self.assertTrue(xmlc.cleaner.is_usable())

        # run preview
        self.run_all(xmlc, False)
        self.assertExists(test_log_path_a)
        self.assertExists(test_log_path_b)

        # really delete
        self.run_all(xmlc, True)
        self.assertNotExists(test_log_path_a)
        self.assertNotExists(test_log_path_b)
